"""Generate C++ code for de/serialization of instances from JSON."""

import io
import itertools
from typing import List, Tuple, Optional, Iterable

from icontract import ensure, require

from aas_core_codegen import intermediate, specific_implementations, naming
from aas_core_codegen.common import (
    Stripped,
    indent_but_first_line,
    Identifier,
    Error,
    assert_never,
)
from aas_core_codegen.cpp import common as cpp_common, naming as cpp_naming
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


def _generate_deserialization_definitions(
    symbol_table: intermediate.SymbolTable,
) -> List[Stripped]:
    """Generate the definitions of deserialization functions."""
    blocks = []  # type: List[Stripped]

    for cls in symbol_table.classes:
        interface_name = cpp_naming.interface_name(cls.name)
        deserialization_name = cpp_naming.function_name(Identifier(f"{cls.name}_from"))

        blocks.append(
            Stripped(
                f"""\
/**
 * \\brief Deserialize \\p json value to an instance
 * of types::{interface_name}.
 *
 * \\param json value to be de-serialized
 * \\param additional_properties if not set, check that \\p json contains
 * no additional properties
 * \\return The deserialized instance, or a de-serialization error, if any.
 */
common::expected<
{I}std::shared_ptr<types::{interface_name}>,
{I}DeserializationError
> {deserialization_name}(
{I}const nlohmann::json& json,
{I}bool additional_properties = false
);"""
            )
        )

    return blocks


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_header(
    symbol_table: intermediate.SymbolTable, library_namespace: Stripped
) -> str:
    """Generate the C++ header code for JSON de/serialization."""
    namespace = Stripped(f"{library_namespace}::{cpp_common.JSONIZATION_NAMESPACE}")

    include_guard_var = cpp_common.include_guard_var(namespace)

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        Stripped(
            f"""\
#ifndef {include_guard_var}
#define {include_guard_var}"""
        ),
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "{include_prefix_path}/common.hpp"
#include "{include_prefix_path}/iteration.hpp"
#include "{include_prefix_path}/types.hpp"

#pragma warning(push, 0)
#include <nlohmann/json.hpp>

#include <memory>
#include <string>
#include <utility>
#pragma warning(pop)"""
        ),
        cpp_common.generate_namespace_opening(library_namespace),
        Stripped(
            f"""\
/**
 * \\defgroup jsonization De/serialize instances from and to JSON.
 * @{{
 */
namespace {cpp_common.JSONIZATION_NAMESPACE} {{"""
        ),
        Stripped(
            f"""\
/**
 * Represent a segment of a JSON path to some value.
 */
class ISegment {{
 public:
{I}/**
{I} * \\brief Convert the segment to a string in a JSON path.
{I} */
{I}virtual std::wstring ToWstring() const = 0;

{I}virtual std::unique_ptr<ISegment> Clone() const = 0;

{I}virtual ~ISegment() = default;
}};  // class ISegment"""
        ),
        Stripped(
            f"""\
/**
 * Represent a property access on a JSON path.
 */
struct PropertySegment : public ISegment {{
{I}/**
{I} * Name of the property in a JSON object
{I} */
{I}std::wstring name;

{I}PropertySegment(
{II}std::wstring a_name
{I});

{I}std::wstring ToWstring() const override;

{I}std::unique_ptr<ISegment> Clone() const override;

{I}~PropertySegment() override = default;
}};  // struct PropertySegment"""
        ),
        Stripped(
            f"""\
/**
 * Represent an index access on a JSON path.
 */
struct IndexSegment : public ISegment {{
{I}/**
{I} * Index of the value in an array.
{I} */
{I}size_t index;

{I}explicit IndexSegment(
{II}size_t an_index
{I});

{I}std::wstring ToWstring() const override;

{I}std::unique_ptr<ISegment> Clone() const override;

{I}~IndexSegment() override = default;
}};  // struct IndexSegment"""
        ),
        Stripped(
            f"""\
/**
 * Represent a JSON path to some value.
 */
struct Path {{
{I}std::deque<std::unique_ptr<ISegment> > segments;

{I}Path();
{I}Path(const Path& other);
{I}Path(Path&& other);
{I}Path& operator=(const Path& other);
{I}Path& operator=(Path&& other);

{I}std::wstring ToWstring() const;
}};  // struct Path"""
        ),
        Stripped("// region De-serialization"),
        Stripped(
            f"""\
/**
 * Represent a de-serialization error.
 */
struct DeserializationError {{
{I}/**
{I} * Human-readable description of the error
{I} */
{I}std::wstring cause;

{I}/**
{I} * Path to the erroneous value
{I} */
{I}Path path;

{I}explicit DeserializationError(std::wstring a_cause);
{I}DeserializationError(std::wstring a_cause, Path a_path);
}};  // struct DeserializationError"""
        ),
        *_generate_deserialization_definitions(symbol_table=symbol_table),
        Stripped("// endregion Deserialization"),
        Stripped("// region Serialization"),
        Stripped(
            f"""\
/**
 * Represent an error in the serialization of an instance to JSON.
 */
class SerializationException : public std::exception {{
 public:
{I}SerializationException(
{II}std::wstring cause,
{II}iteration::Path path
{I});

{I}const char* what() const noexcept override;

{I}const std::wstring& cause() const noexcept;
{I}const iteration::Path& path() const noexcept;

{I}~SerializationException() noexcept override = default;

 private:
{I}const std::wstring cause_;
{I}const iteration::Path path_;
{I}const std::string msg_;
}};  // class SerializationException"""
        ),
        Stripped(
            f"""\
/**
 * \\brief Serialize \\p that instance to a JSON value.
 *
 * \\param that instance to be serialized
 * \\return The corresponding JSON value
 * \\throw \\ref SerializationException if a value within \\p that instance
 * could not be serialized
 */
nlohmann::json Serialize(
{I}const types::IClass& that
);"""
        ),
        Stripped("// endregion Serialization"),
        Stripped(
            f"""\
}}  // namespace {cpp_common.JSONIZATION_NAMESPACE}
/**@}}*/"""
        ),
        cpp_common.generate_namespace_closing(library_namespace),
        cpp_common.WARNING,
        Stripped(f"#endif  // {include_guard_var}"),
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


def _generate_property_segment_implementation() -> List[Stripped]:
    """Generate the implementation of the struct ``PropertySegment``."""
    return [
        Stripped("// region PropertySegment"),
        Stripped(
            f"""\
PropertySegment::PropertySegment(
{I}std::wstring a_name
) :
{I}name(std::move(a_name)) {{
{I}// Intentionally empty.
}}"""
        ),
        Stripped(
            f"""\
std::wstring PropertySegment::ToWstring() const {{
{I}return common::Concat(
{II}L".",
{II}name
{I});
}}"""
        ),
        Stripped(
            f"""\
std::unique_ptr<ISegment> PropertySegment::Clone() const {{
{I}return common::make_unique<PropertySegment>(*this);
}}"""
        ),
        Stripped("// endregion PropertySegment"),
    ]


def _generate_index_segment_implementation() -> List[Stripped]:
    """Generate the implementation of the struct ``IndexSegment``."""
    return [
        Stripped("// region IndexSegment"),
        Stripped(
            f"""\
IndexSegment::IndexSegment(
{I}size_t an_index
) :
{I}index(an_index) {{
{I}// Intentionally empty.
}}"""
        ),
        Stripped(
            f"""\
std::wstring IndexSegment::ToWstring() const {{
{I}return common::Concat(
{II}L"[",
{II}std::to_wstring(index),
{II}L"]"
{I});
}}"""
        ),
        Stripped(
            f"""\
std::unique_ptr<ISegment> IndexSegment::Clone() const {{
{I}return common::make_unique<IndexSegment>(*this);
}}"""
        ),
        Stripped("// endregion IndexSegment"),
    ]


def _generate_path_implementation() -> List[Stripped]:
    """Generate the implementation of the struct ``Path``."""
    return [
        Stripped("// region struct Path"),
        Stripped(
            f"""\
Path::Path() {{
{I}// Intentionally empty.
}}"""
        ),
        Stripped(
            f"""\
Path::Path(const Path& other) {{
{I}for (const std::unique_ptr<ISegment>& segment : other.segments) {{
{II}segments.emplace_back(segment->Clone());
{I}}}
}}"""
        ),
        Stripped(
            f"""\
Path::Path(Path&& other) {{
{I}segments = std::move(other.segments);
}}"""
        ),
        Stripped(
            f"""\
Path& Path::operator=(const Path& other) {{
{I}segments.clear();
{I}for (const std::unique_ptr<ISegment>& segment : other.segments) {{
{II}segments.emplace_back(segment->Clone());
{I}}}
{I}return *this;
}}"""
        ),
        Stripped(
            f"""\
Path& Path::operator=(Path&& other) {{
{I}if (this != &other) {{
{II}segments = std::move(other.segments);
{I}}}
{I}return *this;
}}"""
        ),
        Stripped(
            f"""\
std::wstring Path::ToWstring() const {{
{I}if (segments.empty()) {{
{II}return L"";
{I}}}

{I}std::deque<std::wstring> parts;
{I}for (const std::unique_ptr<ISegment>& segment : segments ) {{
{II}parts.emplace_back(segment->ToWstring());
{I}}}

{I}size_t out_len = 0;
{I}for (const std::wstring& part : parts) {{
{II}out_len += part.size();
{I}}}

{I}std::wstring out;
{I}out.reserve(out_len);
{I}for (const std::wstring& part : parts) {{
{II}out.append(part);
{I}}}

{I}return out;
}}"""
        ),
        Stripped("// endregion struct Path"),
    ]


def _generate_deserialization_error_implementation() -> List[Stripped]:
    """Generate the impl. of the deserialization error class."""
    return [
        Stripped("// region class DeserializationError"),
        Stripped(
            f"""\
DeserializationError::DeserializationError(
{I}std::wstring a_cause
) :
{I}cause(a_cause) {{
{I}// Intentionally empty.
}}"""
        ),
        Stripped(
            f"""\
DeserializationError::DeserializationError(
{I}std::wstring a_cause,
{I}Path a_path
) :
{I}cause(a_cause),
{I}path(a_path) {{
{I}// Intentionally empty.
}}"""
        ),
        Stripped("// endregion class DeserializationError"),
    ]


def _generate_deserialize_bool() -> Stripped:
    """Generate the function to de-serialize a boolean from a JSON value."""
    return Stripped(
        f"""\
std::pair<
{I}common::optional<bool>,
{I}common::optional<DeserializationError>
> DeserializeBool(
{I}const nlohmann::json& json
) {{
{I}if (!json.is_boolean()) {{
{II}std::wstring message = common::Concat(
{III}L"Expected a boolean, but got a value of type: ",
{III}common::Utf8ToWstring(json.type_name())
{II});

{II}return std::make_pair<
{III}common::optional<bool>,
{III}common::optional<DeserializationError>
{II}>(
{III}common::nullopt,
{III}common::make_optional<DeserializationError>(
{IIII}message
{III})
{II});
{I}}}

{I}return std::make_pair<
{II}common::optional<bool>,
{II}common::optional<DeserializationError>
{I}>(
{II}json.get<bool>(),
{II}common::nullopt
{I});
}}"""
    )


def _generate_deserialize_int() -> Stripped:
    """Generate the function to de-serialize an int from a JSON value."""
    return Stripped(
        f"""\
std::pair<
{I}common::optional<int64_t>,
{I}common::optional<DeserializationError>
> DeserializeInt64(
{I}const nlohmann::json& json
) {{
{I}if (!json.is_number()) {{
{II}std::wstring message = common::Concat(
{III}L"Expected an integer number, but got a value of type: ",
{III}common::Utf8ToWstring(json.type_name())
{II});

{II}return std::make_pair<
{III}common::optional<int64_t>,
{III}common::optional<DeserializationError>
{II}>(
{III}common::nullopt,
{III}common::make_optional<DeserializationError>(
{IIII}message
{III})
{II});
{I}}}

{I}static_assert(
{II}std::is_same<nlohmann::json::number_integer_t, int64_t>::value,
{II}"Expected nlohmann::json::number_integer_t to equal int64_t, "
{II}"but it does not."
{I});

{I}if (json.is_number_integer()) {{
{II}return std::make_pair<
{III}common::optional<int64_t>,
{III}common::optional<DeserializationError>
{II}>(
{III}json.get<int64_t>(),
{III}common::nullopt
{II});
{I}}}

{I}if (json.is_number_unsigned()) {{
{II}std::wstring message = common::Concat(
{III}L"Expected a 64-bit integer number, "
{III}L"but got an unsigned integer number which does not fit in that range: ",
{III}std::to_wstring(json.get<nlohmann::json::number_unsigned_t>())
{II});

{II}return std::make_pair<
{III}common::optional<int64_t>,
{III}common::optional<DeserializationError>
{II}>(
{III}common::nullopt,
{III}common::make_optional<DeserializationError>(
{IIII}message
{III})
{II});
{I}}}

{I}// NOTE (mristin):
{I}// We have to check that the number is an integer even though it can
{I}// not be stored in int64_t in order to give an informative message.

{I}const nlohmann::json::number_float_t number(
{II}json.get<nlohmann::json::number_float_t>()
{I});

{I}nlohmann::json::number_float_t integer_part;
{I}const bool is_integer(
{II}std::modf(number, &integer_part) == 0
{I});
{I}if (is_integer) {{
{II}std::wstring message = common::Concat(
{III}L"Expected a 64-bit integer number, "
{III}L"but got an integer number which does not fit in that range: ",
{III}std::to_wstring(number)
{II});

{II}return std::make_pair<
{III}common::optional<int64_t>,
{III}common::optional<DeserializationError>
{II}>(
{III}common::nullopt,
{III}common::make_optional<DeserializationError>(
{IIII}message
{III})
{II});
{I}}} else {{
{II}std::wstring message = common::Concat(
{III}L"Expected a 64-bit integer number, "
{III}L"but got a non-integer number: ",
{III}std::to_wstring(number)
{II});

{II}return std::make_pair<
{III}common::optional<int64_t>,
{III}common::optional<DeserializationError>
{II}>(
{III}common::nullopt,
{III}common::make_optional<DeserializationError>(
{IIII}message
{III})
{II});
{I}}}
}}"""
    )


def _generate_deserialize_float() -> Stripped:
    """Generate the function to de-serialize a float from a JSON value."""
    return Stripped(
        f"""\
std::pair<
{I}common::optional<double>,
{I}common::optional<DeserializationError>
> DeserializeDouble(
{I}const nlohmann::json& json
) {{
{I}if (!json.is_number()) {{
{II}std::wstring message = common::Concat(
{III}L"Expected a number, but got a value of type: ",
{III}common::Utf8ToWstring(json.type_name())
{II});

{II}return std::make_pair<
{III}common::optional<double>,
{III}common::optional<DeserializationError>
{II}>(
{III}common::nullopt,
{III}common::make_optional<DeserializationError>(
{IIII}message
{III})
{II});
{I}}}

{I}static_assert(
{II}std::is_same<nlohmann::json::number_float_t, double>::value,
{II}"Expected nlohmann::json::number_float_t to equal double, "
{II}"but it does not."
{I});

{I}return std::make_pair<
{II}common::optional<double>,
{II}common::optional<DeserializationError>
{I}>(
{II}json.get<double>(),
{II}common::nullopt
{I});
}}"""
    )


def _generate_deserialize_str() -> Stripped:
    """Generate the function to de-serialize a string from a JSON value."""
    return Stripped(
        f"""\
std::pair<
{I}common::optional<std::wstring>,
{I}common::optional<DeserializationError>
> DeserializeWstring(
{I}const nlohmann::json& json
) {{
{I}if (!json.is_string()) {{
{II}std::wstring message = common::Concat(
{III}L"Expected a string, but got a value of type: ",
{III}common::Utf8ToWstring(json.type_name())
{II});

{II}return std::make_pair<
{III}common::optional<std::wstring>,
{III}common::optional<DeserializationError>
{II}>(
{III}common::nullopt,
{III}common::make_optional<DeserializationError>(
{IIII}message
{III})
{II});
{I}}}

{I}return std::make_pair<
{II}common::optional<std::wstring>,
{II}common::optional<DeserializationError>
{I}>(
{II}common::Utf8ToWstring(*(json.get_ptr<const std::string*>())),
{II}common::nullopt
{I});
}}"""
    )


def _generate_deserialize_bytearray() -> Stripped:
    """Generate the function to de-serialize a bytearray from a JSON value."""
    return Stripped(
        f"""\
std::pair<
{I}common::optional<std::vector<std::uint8_t> >,
{I}common::optional<DeserializationError>
> DeserializeByteArray(
{I}const nlohmann::json& json
) {{
{I}if (!json.is_string()) {{
{II}std::wstring message = common::Concat(
{III}L"Expected a base64-encoded byte array as a string, "
{III}L"but got a value of type: ",
{III}common::Utf8ToWstring(json.type_name())
{II});;

{II}return std::make_pair<
{III}common::optional<std::vector<std::uint8_t> >,
{III}common::optional<DeserializationError>
{II}>(
{III}common::nullopt,
{III}common::make_optional<DeserializationError>(
{IIII}message
{III})
{II});
{I}}}

{I}common::expected<
{II}std::vector<std::uint8_t>,
{II}std::string
{I}> bytes = stringification::Base64Decode(
{II}*(json.get_ptr<const std::string*>())
{I});

{I}if (!bytes.has_value()) {{
{II}std::wstring message = common::Concat(
{III}L"Failed to base64-decode the bytes from a string: ",
{III}common::Utf8ToWstring(bytes.error())
{II});

{II}return std::make_pair<
{III}common::optional<std::vector<std::uint8_t> >,
{III}common::optional<DeserializationError>
{II}>(
{III}common::nullopt,
{III}common::make_optional<DeserializationError>(
{IIII}message
{III})
{II});
{I}}}

{I}return std::make_pair<
{II}common::optional<std::vector<std::uint8_t> >,
{II}common::optional<DeserializationError>
{I}>(
{II}std::move(*bytes),
{II}common::nullopt
{I});
}}"""
    )


_PRIMITIVE_TYPE_TO_DESERIALIZE = {
    intermediate.PrimitiveType.BOOL: "DeserializeBool",
    intermediate.PrimitiveType.INT: "DeserializeInt64",
    intermediate.PrimitiveType.FLOAT: "DeserializeDouble",
    intermediate.PrimitiveType.STR: "DeserializeWstring",
    intermediate.PrimitiveType.BYTEARRAY: "DeserializeByteArray",
}
assert all(
    primitive_type in _PRIMITIVE_TYPE_TO_DESERIALIZE
    for primitive_type in intermediate.PrimitiveType
)


def _generate_get_model_type() -> Stripped:
    """Generate the getter of the model type from JSON object for dispatches."""
    return Stripped(
        f"""\
/**
 * Get the property `modelType` from the JSON value expected as a JSON object.
 */
std::pair<
{I}const std::string*,
{I}common::optional<DeserializationError>
> GetModelTypeFrom(
{I}const nlohmann::json& json
) {{
{I}if (!json.is_object()) {{
{II}std::wstring message = common::Concat(
{III}L"Expected an object, but got: ",
{III}common::Utf8ToWstring(json.type_name())
{II});

{II}return std::make_pair<
{III}const std::string*,
{III}common::optional<DeserializationError>
{II}>(
{III}nullptr,
{III}common::make_optional<DeserializationError>(
{IIII}message
{III})
{II});
{I}}}

{I}if (!json.contains("modelType")) {{
{II}return std::make_pair<
{III}const std::string*,
{III}common::optional<DeserializationError>
{II}>(
{III}nullptr,
{III}common::make_optional<DeserializationError>(
{IIII}L"The required property modelType is missing"
{III})
{II});
{I}}}

{I}const nlohmann::json& model_type_prop = json["modelType"];
{I}if (!model_type_prop.is_string()) {{
{II}std::wstring message = common::Concat(
{III}L"Expected modelType to be a string, but got: ",
{III}common::Utf8ToWstring(model_type_prop.type_name())
{II});

{II}common::optional<DeserializationError> error(
{III}common::make_optional<DeserializationError>(
{IIII}message
{III})
{II});
{II}error->path.segments.emplace_front(
{III}common::make_unique<PropertySegment>(
{IIII}L"modelType"
{III})
{II});

{II}// NOTE (mristin):
{II}// We have to explicitly use the constructor instead of std::make_pair
{II}// as `const std::string*` can not be automatically converted to a rvalue.
{II}return std::pair<
{III}const std::string*,
{III}common::optional<DeserializationError>
{II}>(
{III}nullptr,
{III}error
{II});
{I}}}

{I}static_assert(
{II}std::is_same<nlohmann::json::string_t, std::string>::value,
{II}"Expected nlohmann::json::string_t to equal std::string, but it does not."
{I});

{I}const std::string* model_type(
{II}model_type_prop.get_ptr<const std::string*>()
{I});

{I}// NOTE (mristin):
{I}// We have to explicitly use the constructor instead of std::make_pair
{I}// as `const std::string*` can not be automatically converted to a rvalue.
{I}return std::pair<
{II}const std::string*,
{II}common::optional<DeserializationError>
{I}>(
{II}model_type,
{II}common::nullopt
{I});
}}"""
    )


def _generate_concretely_deserialize_definition(
    cls: intermediate.ClassUnion,
) -> Stripped:
    """Generate the definition of the concrete ``Deserialize*`` functions."""
    interface_name = cpp_naming.interface_name(cls.name)

    if len(cls.concrete_descendants) == 0:
        function_name = cpp_naming.function_name(Identifier(f"deserialize_{cls.name}"))
    else:
        function_name = cpp_naming.function_name(
            Identifier(f"concretely_deserialize_{cls.name}")
        )

    if len(cls.ancestors) > 0:
        # NOTE (mristin, 2023-11-10):
        # We have to introduce the template so that we do not have to
        # unnecessarily upcast the instance to ancestor classes.
        prefix = Stripped(
            f"""\
template <
{I}typename T,
{I}typename std::enable_if<
{II}std::is_base_of<T, types::{interface_name}>::value
{I}>::type* = nullptr
>
std::pair<
{I}common::optional<std::shared_ptr<T> >,
{I}common::optional<DeserializationError>
>"""
        )
    else:
        prefix = Stripped(
            f"""\
std::pair<
{I}common::optional<
{II}std::shared_ptr<types::{interface_name}>
{I}>,
{I}common::optional<DeserializationError>
>"""
        )

    return Stripped(
        f"""\
/**
 * \\brief Deserialize concretely an instance
 * of types::{interface_name}.
 *
 * \\param json value to be de-serialized
 * \\param additional_properties if not set, check that \\p json contains
 * no additional properties
 * \\return the deserialized instance, or an error, if any
 */
{prefix} {function_name}(
{I}const nlohmann::json& json,
{I}bool additional_properties
);"""
    )


@require(
    lambda cls: len(cls.concrete_descendants) > 0,
    "No dispatch possible without concrete descendants",
)
def _generate_dispatch_deserialize_definition(cls: intermediate.ClassUnion) -> Stripped:
    """Generate the def. of the dispatching deserialization function for ``cls``."""
    interface_name = cpp_naming.interface_name(cls.name)

    function_name = cpp_naming.function_name(Identifier(f"deserialize_{cls.name}"))

    return Stripped(
        f"""\
/**
 * \\brief Dispatch the deserialization for an instance
 * of types::{interface_name}.
 *
 * \\param json value to be de-serialized
 * \\param additional_properties if not set, check that \\p json contains
 * no additional properties
 * \\return the deserialized instance, or an error, if any
 */
std::pair<
{I}common::optional<
{II}std::shared_ptr<types::{interface_name}>
{I}>,
{I}common::optional<DeserializationError>
> {function_name}(
{I}const nlohmann::json& json,
{I}bool additional_properties
);"""
    )


def _generate_deserialize_primitive_property(
    prop: intermediate.Property, ok_type: Stripped
) -> Stripped:
    """
    Generate the snippet to de-serialize the primitive property.

    We assume that the check whether the property is set is performed elsewhere.

    The ``ok_type`` denotes the type of the deserialized instance, *not* the property.
    We have to distinguish between cases where we directly create an upcast pointer to
    an ancestor class, and cases where there are no ancestor classes.
    """
    type_anno = intermediate.beneath_optional(prop.type_annotation)
    primitive_type = intermediate.try_primitive_type(type_anno)

    assert (
        primitive_type is not None
    ), f"Primitive property expected, got for {prop.name!r}: {prop.type_annotation}"

    deserialize_primitive = _PRIMITIVE_TYPE_TO_DESERIALIZE[primitive_type]

    json_prop_name = naming.json_property(prop.name)

    var_name = cpp_naming.variable_name(Identifier(f"the_{prop.name}"))

    return Stripped(
        f"""\
std::tie(
{I}{var_name},
{I}error
) = {deserialize_primitive}(
{I}json[{cpp_common.string_literal(json_prop_name)}]
);

if (error.has_value()) {{
{I}error->path.segments.emplace_front(
{II}common::make_unique<PropertySegment>(
{III}{cpp_common.wstring_literal(json_prop_name)}
{II})
{I});

{I}return std::make_pair<
{II}common::optional<std::shared_ptr<{ok_type}> >,
{II}common::optional<DeserializationError>
{I}>(
{II}common::nullopt,
{II}std::move(error)
{I});
}}"""
    )


def _generate_deserialize_enumeration_property(
    prop: intermediate.Property, ok_type: Stripped
) -> Stripped:
    """
    Generate the snippet to de-serialize the enumeration property.

    We assume that the check whether the property is set is performed elsewhere.

    The ``ok_type`` denotes the type of the deserialized instance, *not* the property.
    We have to distinguish between cases where we directly create an upcast pointer to
    an ancestor class, and cases where there are no ancestor classes.
    """
    type_anno = intermediate.beneath_optional(prop.type_annotation)
    assert isinstance(type_anno, intermediate.OurTypeAnnotation) and isinstance(
        type_anno.our_type, intermediate.Enumeration
    )

    enum = type_anno.our_type
    enum_name = cpp_naming.enum_name(enum.name)

    from_wstring = cpp_naming.function_name(Identifier(f"{enum.name}_from_wstring"))

    json_prop_name = naming.json_property(prop.name)

    var_name = cpp_naming.variable_name(Identifier(f"the_{prop.name}"))
    var_text_name = cpp_naming.variable_name(Identifier(f"text_{prop.name}"))

    return Stripped(
        f"""\
common::optional<std::wstring> {var_text_name};

std::tie(
{I}{var_text_name},
{I}error
) = DeserializeWstring(
{I}json[{cpp_common.string_literal(json_prop_name)}]
);

if (error.has_value()) {{
{I}error->path.segments.emplace_front(
{II}common::make_unique<PropertySegment>(
{III}{cpp_common.wstring_literal(json_prop_name)}
{II})
{I});

{I}return std::make_pair<
{II}common::optional<std::shared_ptr<{ok_type}> >,
{II}common::optional<DeserializationError>
{I}>(
{II}common::nullopt,
{II}std::move(error)
{I});
}}

{var_name} = std::move(
{I}wstringification::{from_wstring}(
{II}*{var_text_name}
{I})
);
if (!{var_name}.has_value()) {{
{I}std::wstring message = common::Concat(
{II}L"Invalid literal for {enum_name}: ",
{II}*{var_text_name}
{I});

{I}error = common::make_optional<DeserializationError>(
{II}message
{I});
{I}error->path.segments.emplace_front(
{II}common::make_unique<PropertySegment>(
{III}{cpp_common.wstring_literal(json_prop_name)}
{II})
{I});

{I}return std::make_pair<
{II}common::optional<std::shared_ptr<{ok_type}> >,
{II}common::optional<DeserializationError>
{I}>(
{II}common::nullopt,
{II}std::move(error)
{I});
}}"""
    )


def _determine_deserialize_function_to_call(cls: intermediate.ClassUnion) -> Stripped:
    """
    Determine the function to be called to de-serialize an instance of ``cls``.

    The result includes also template parameters, if any are necessary.
    """
    deserialize_name = cpp_naming.function_name(Identifier(f"Deserialize_{cls.name}"))

    interface_name = cpp_naming.interface_name(cls.name)

    if len(cls.concrete_descendants) == 0:
        if len(cls.ancestors) == 0:
            deserialize_function = Stripped(deserialize_name)
        else:
            deserialize_function = Stripped(
                f"""\
{deserialize_name}<
{I}types::{interface_name}
>"""
            )
    else:
        deserialize_function = Stripped(deserialize_name)

    return deserialize_function


def _generate_deserialize_instance_property(
    prop: intermediate.Property, ok_type: Stripped
) -> Stripped:
    """
    Generate the snippet to de-serialize the instance property.

    We assume that the check whether the property is set is performed elsewhere.

    The ``ok_type`` denotes the type of the deserialized instance, *not* the property.
    We have to distinguish between cases where we directly create an upcast pointer to
    an ancestor class, and cases where there are no ancestor classes.
    """
    type_anno = intermediate.beneath_optional(prop.type_annotation)
    assert isinstance(type_anno, intermediate.OurTypeAnnotation) and isinstance(
        type_anno.our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
    )

    cls = type_anno.our_type

    deserialize_function = _determine_deserialize_function_to_call(cls=cls)

    var_name = cpp_naming.variable_name(Identifier(f"the_{prop.name}"))
    json_prop_name = naming.json_property(prop.name)

    return Stripped(
        f"""\
std::tie(
{I}{var_name},
{I}error
) = {deserialize_function}(
{I}json[{cpp_common.string_literal(json_prop_name)}],
{I}additional_properties
);

if (error.has_value()) {{
{I}error->path.segments.emplace_front(
{II}common::make_unique<PropertySegment>(
{III}{cpp_common.wstring_literal(json_prop_name)}
{II})
{I});

{I}return std::make_pair<
{II}common::optional<std::shared_ptr<{ok_type}> >,
{II}common::optional<DeserializationError>
{I}>(
{II}common::nullopt,
{II}std::move(error)
{I});
}}"""
    )


def _generate_deserialize_list_property(
    prop: intermediate.Property, ok_type: Stripped
) -> Stripped:
    """
    Generate the snippet to de-serialize the list property.

    We assume that the check whether the property is set is performed elsewhere.

    The ``ok_type`` denotes the type of the deserialized instance, *not* the property.
    We have to distinguish between cases where we directly create an upcast pointer to
    an ancestor class, and cases where there are no ancestor classes.
    """
    type_anno = intermediate.beneath_optional(prop.type_annotation)
    assert isinstance(type_anno, intermediate.ListTypeAnnotation)
    assert isinstance(type_anno.items, intermediate.OurTypeAnnotation) and isinstance(
        type_anno.items.our_type,
        (intermediate.AbstractClass, intermediate.ConcreteClass),
    ), (
        f"NOTE (mristin, 2023-11-10): We expect only lists of classes "
        f"at the moment, but you specified {type_anno}. "
        f"Please contact the developers if you need this feature."
    )

    cls = type_anno.items.our_type

    interface_name = cpp_naming.interface_name(cls.name)

    deserialize_function = _determine_deserialize_function_to_call(cls=cls)

    var_name = cpp_naming.variable_name(Identifier(f"the_{prop.name}"))
    json_prop_name = naming.json_property(prop.name)

    var_index_name = cpp_naming.variable_name(Identifier(f"index_{prop.name}"))
    var_json = cpp_naming.variable_name(Identifier(f"json_{prop.name}"))

    return Stripped(
        f"""\
const nlohmann::json& {var_json}(
{I}json[{cpp_common.string_literal(json_prop_name)}]
);
if (!{var_json}.is_array()) {{
{I}error = common::make_optional<DeserializationError>(
{II}common::Concat(
{III}L"Expected an array, but got: ",
{III}common::Utf8ToWstring(
{IIII}{var_json}.type_name()
{III})
{II})
{I});

{I}error->path.segments.emplace_front(
{II}common::make_unique<PropertySegment>(
{III}{cpp_common.wstring_literal(json_prop_name)}
{II})
{I});

{I}return std::make_pair<
{II}common::optional<std::shared_ptr<{ok_type}> >,
{II}common::optional<DeserializationError>
{I}>(
{II}common::nullopt,
{II}std::move(error)
{I});
}}

{var_name} = common::make_optional<
{I}std::vector<
{II}std::shared_ptr<types::{interface_name}>
{I}>
>();

{var_name}->reserve({var_json}.size());

size_t {var_index_name} = 0;

for (
{I}const nlohmann::json& item
{I}: {var_json}
) {{
{I}common::optional<
{II}std::shared_ptr<types::{interface_name}>
{I}> deserialized;

{I}std::tie(
{II}deserialized,
{II}error
{I}) = {indent_but_first_line(deserialize_function, I)}(
{II}item,
{II}additional_properties
{I});

{I}if (error.has_value()) {{
{II}error->path.segments.emplace_front(
{III}common::make_unique<IndexSegment>(
{IIII}{var_index_name}
{III})
{II});

{II}error->path.segments.emplace_front(
{III}common::make_unique<PropertySegment>(
{IIII}{cpp_common.wstring_literal(json_prop_name)}
{III})
{II});

{II}return std::make_pair<
{III}common::optional<std::shared_ptr<{ok_type}> >,
{III}common::optional<DeserializationError>
{II}>(
{III}common::nullopt,
{III}std::move(error)
{II});
{I}}}

{I}{var_name}->emplace_back(
{II}std::move(*deserialized)
{I});

{I}++{var_index_name};
}}"""
    )


def _generate_deserialize_property(
    prop: intermediate.Property, ok_type: Stripped
) -> Stripped:
    """
    Generate the snippet to de-serialize the given property.

    The ``ok_type`` denotes the type of the deserialized instance, *not* the property.
    We have to distinguish between cases where we directly create an upcast pointer to
    an ancestor class, and cases where there are no ancestor classes.
    """
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    code: Stripped

    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
        code = _generate_deserialize_primitive_property(prop=prop, ok_type=ok_type)

    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        if isinstance(type_anno.our_type, intermediate.Enumeration):
            code = _generate_deserialize_enumeration_property(
                prop=prop, ok_type=ok_type
            )

        elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
            code = _generate_deserialize_primitive_property(prop=prop, ok_type=ok_type)

        elif isinstance(
            type_anno.our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            code = _generate_deserialize_instance_property(prop=prop, ok_type=ok_type)

        else:
            assert_never(type_anno.our_type)
    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        assert isinstance(
            type_anno.items, intermediate.OurTypeAnnotation
        ) and isinstance(
            type_anno.items.our_type,
            (intermediate.AbstractClass, intermediate.ConcreteClass),
        ), (
            f"NOTE (mristin, 2023-11-10): We expect only lists of classes "
            f"at the moment, but you specified {type_anno}. "
            f"Please contact the developers if you need this feature."
        )

        code = _generate_deserialize_list_property(prop=prop, ok_type=ok_type)
    else:
        assert_never(type_anno)

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        json_prop_literal = cpp_common.string_literal(naming.json_property(prop.name))

        code = Stripped(
            f"""\
if (json.contains({json_prop_literal})) {{
{I}{indent_but_first_line(code, I)}
}}"""
        )

    return code


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_concretely_deserialize_implementation(
    cls: intermediate.ConcreteClass,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """
    Generate the concrete deserialization for the class ``cls``.

    It is assumed that the dispatch has been already effectuated to this generated
    function, so no further dispatch should be performed.
    """
    if cls.is_implementation_specific:
        implementation_key = specific_implementations.ImplementationKey(
            f"jsonization/deserialize_{cls.name}.cpp"
        )

        code = spec_impls.get(implementation_key, None)
        if code is None:
            return None, Error(
                cls.parsed.node,
                f"The implementation is missing for the JSON deserialization "
                f"of {cls.name!r}: {implementation_key}",
            )
        return code, None

    interface_name = cpp_naming.interface_name(cls.name)

    if len(cls.ancestors) == 0:
        # NOTE (mristin, 2023-11-10):
        # We will not need to upcast this instance, so the return value type
        # is simply the interface.
        ok_type = Stripped(f"types::{interface_name}")
    else:
        # NOTE (mristin, 2023-11-10):
        # We have to leave it open to upcast to an ancestor class, and hence
        # we introduce a template parameter.
        ok_type = Stripped("T")

    expected_properties = cpp_naming.constant_name(
        Identifier(f"properties_in_{cls.name}")
    )

    blocks = [
        Stripped(
            f"""\
if (!json.is_object()) {{
{I}std::wstring message = common::Concat(
{II}L"Expected an object, but got: ",
{II}common::Utf8ToWstring(json.type_name())
{I});

{I}return std::make_pair<
{II}common::optional<std::shared_ptr<{ok_type}> >,
{II}common::optional<DeserializationError>
{I}>(
{II}common::nullopt,
{II}common::make_optional<DeserializationError>(
{III}message
{II})
{I});
}}"""
        ),
        Stripped(
            f"""\
if (!additional_properties) {{
{I}for (const auto& key_val : json.items()) {{
{II}auto it(
{III}{expected_properties}.find(key_val.key())
{II});
{II}if (it == {expected_properties}.end()) {{
{III}std::wstring message = common::Concat(
{IIII}L"Unexpected additional property: ",
{IIII}common::Utf8ToWstring(key_val.key())
{III});

{III}return std::make_pair<
{IIII}common::optional<std::shared_ptr<{ok_type}> >,
{IIII}common::optional<DeserializationError>
{III}>(
{IIII}common::nullopt,
{IIII}common::make_optional<DeserializationError>(
{IIIII}message
{IIII})
{III});
{II}}}
{I}}}
}}"""
        ),
    ]  # type: List[Stripped]

    class_name = cpp_naming.class_name(cls.name)

    if len(cls.properties) == 0:
        blocks.append(
            Stripped(
                f"""\
return std::make_pair(
{I}common::make_optional<
{II}std::shared_ptr<{ok_type}>
{I}>(
{II}// NOTE (mristin):
{II}// We deliberately do not use std::make_shared here to avoid an unnecessary
{II}// upcast.
{II}new types::{class_name}()
{I}),
{I}common::nullopt
);"""
            )
        )
    else:
        names_of_required_properties = [
            prop.name
            for prop in cls.properties
            if not isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation)
        ]

        if cls.serialization.with_model_type:
            names_of_required_properties.append(Identifier("model_type"))

        if len(names_of_required_properties) > 0:
            blocks.append(Stripped("// region Check required properties"))
            for prop_name in names_of_required_properties:
                json_prop_name = naming.json_property(prop_name)
                json_prop_name_literal = cpp_common.string_literal(json_prop_name)
                blocks.append(
                    Stripped(
                        f"""\
if (!json.contains({json_prop_name_literal})) {{
{I}return std::make_pair<
{II}common::optional<std::shared_ptr<{ok_type}> >,
{II}common::optional<DeserializationError>
{I}>(
{II}common::nullopt,
{II}common::make_optional<DeserializationError>(
{III}L"The required property {json_prop_name} is missing"
{II})
{I});
}}"""
                    )
                )

            blocks.append(Stripped("// endregion Check required properties"))

        # region Initialization
        blocks.append(Stripped("// region Initialization"))

        if len(cls.properties) > 0:
            blocks.append(Stripped("common::optional<DeserializationError> error;"))

        init_statements = []  # type: List[Stripped]

        for prop in cls.properties:
            var_name = cpp_naming.variable_name(Identifier(f"the_{prop.name}"))
            var_type = cpp_common.generate_type(
                type_annotation=prop.type_annotation,
                types_namespace=cpp_common.TYPES_NAMESPACE,
            )

            if not isinstance(
                prop.type_annotation, intermediate.OptionalTypeAnnotation
            ):
                if "\n" in var_type:
                    var_type = Stripped(
                        f"""\
common::optional<
{I}{indent_but_first_line(var_type, I)}
>"""
                    )
                else:
                    if var_type.endswith(">"):
                        var_type = Stripped(f"common::optional<{var_type} >")
                    else:
                        var_type = Stripped(f"common::optional<{var_type}>")

            init_statements.append(Stripped(f"{var_type} {var_name};"))

        blocks.append(Stripped("\n\n".join(init_statements)))

        blocks.append(Stripped("// endregion Initialization"))
        # endregion

        # region Deserialize properties
        for prop in cls.properties:
            json_prop_name = naming.json_property(prop.name)
            blocks.append(Stripped(f"// region De-serialize {json_prop_name}"))

            blocks.append(_generate_deserialize_property(prop=prop, ok_type=ok_type))

            blocks.append(Stripped(f"// endregion De-serialize {json_prop_name}"))
        # endregion

        if cls.serialization.with_model_type:
            # NOTE (mristin):
            # If the serialization requires a model type, we consequently check for it
            # here. The model type thus obtained is *not* used for any dispatch. We only
            # use this value for verification to make sure that the model type
            # of the instances is consistent with the expected value for its concrete
            # class. This will be performed even though the code might have had to parse
            # model type before for the dispatch. We decided to double-check to cover
            # the case where a dispatch is *unnecessary* (*e.g.*, the caller knows the
            # expected runtime type), but the model type might still be invalid in the
            # input. Hence, when the dispatch is *necessary*, the model type JSON
            # property will be parsed twice, which is a cost we currently find
            # acceptable.

            blocks.append(
                Stripped(
                    """\
// region Check model type
// This check is intended only for verification, not for dispatch."""
                )
            )

            model_type = naming.json_model_type(cls.name)

            blocks.append(
                Stripped(
                    f"""\
common::optional<
{I}std::wstring
> model_type;

std::tie(
{I}model_type,
{I}error
) = DeserializeWstring(
{I}json["modelType"]
);

if (error.has_value()) {{
{I}error->path.segments.emplace_front(
{II}common::make_unique<PropertySegment>(
{III}L"modelType"
{II})
{I});

{I}return std::make_pair<
{II}common::optional<std::shared_ptr<{ok_type}> >,
{II}common::optional<DeserializationError>
{I}>(
{II}common::nullopt,
{II}std::move(error)
{I});
}}

if (*model_type != L"{model_type}") {{
{I}std::wstring message = common::Concat(
{II}L"Expected model type '{model_type}', "
{II}L"but got: ",
{II}*model_type
{I});

{I}error = common::make_optional<DeserializationError>(
{II}message
{I});

{I}return std::make_pair<
{II}common::optional<std::shared_ptr<{ok_type}> >,
{II}common::optional<DeserializationError>
{I}>(
{II}common::nullopt,
{II}std::move(error)
{I});
}}"""
                )
            )

            blocks.append(Stripped("// endregion Check model type"))

        # region Pass arguments to the constructor
        property_names = [prop.name for prop in cls.properties]
        constructor_argument_names = [arg.name for arg in cls.constructor.arguments]

        # fmt: off
        assert (
                set(prop.name for prop in cls.properties)
                == set(arg.name for arg in cls.constructor.arguments)
        ), (
            f"Expected the properties to coincide with constructor arguments, "
            f"but they do not for {cls.name!r}:"
            f"{property_names=}, {constructor_argument_names=}"
        )
        # fmt: on

        constructor_args = []  # type: List[Stripped]
        for arg in cls.constructor.arguments:
            prop = cls.properties_by_name[arg.name]

            var_name = cpp_naming.variable_name(Identifier(f"the_{prop.name}"))
            if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
                constructor_args.append(Stripped(f"std::move({var_name})"))
            else:
                constructor_args.append(Stripped(f"std::move(*{var_name})"))

        constructor_args_joined = ",\n".join(constructor_args)

        blocks.append(
            Stripped(
                f"""\
return std::make_pair(
{I}common::make_optional<
{II}std::shared_ptr<{ok_type}>
{I}>(
{II}// NOTE (mristin):
{II}// We deliberately do not use std::make_shared here to avoid an unnecessary
{II}// upcast.
{II}new types::{class_name}(
{III}{indent_but_first_line(constructor_args_joined, III)}
{II})
{I}),
{I}common::nullopt
);"""
            )
        )
        # endregion

    body = "\n\n".join(blocks)

    if len(cls.concrete_descendants) == 0:
        function_name = cpp_naming.function_name(Identifier(f"deserialize_{cls.name}"))
    else:
        function_name = cpp_naming.function_name(
            Identifier(f"concretely_deserialize_{cls.name}")
        )

    if len(cls.ancestors) > 0:
        # NOTE (mristin, 2023-11-10):
        # We have to introduce the template so that we do not have to
        # unnecessarily upcast the instance to ancestor classes.
        prefix = Stripped(
            f"""\
template <
{I}typename T,
{I}typename std::enable_if<
{II}std::is_base_of<T, types::{interface_name}>::value
{I}>::type*
>
std::pair<
{I}common::optional<std::shared_ptr<T> >,
{I}common::optional<DeserializationError>
>"""
        )
    else:
        prefix = Stripped(
            f"""\
std::pair<
{I}common::optional<
{II}std::shared_ptr<types::{interface_name}>
{I}>,
{I}common::optional<DeserializationError>
>"""
        )

    expected_properties_literals = [
        f"{cpp_common.string_literal(naming.json_property(prop.name))}"
        for prop in cls.properties
    ]

    if cls.serialization.with_model_type:
        expected_properties_literals.append(cpp_common.string_literal("modelType"))

    expected_properties_literals_joined = ",\n".join(expected_properties_literals)

    expected_properties_definition = Stripped(
        f"""\
std::set<std::string> {expected_properties} = {{
{I}{indent_but_first_line(expected_properties_literals_joined, I)}
}};"""
    )

    return (
        Stripped(
            f"""\
{expected_properties_definition}

{prefix} {function_name}(
{I}const nlohmann::json& json,
{I}bool additional_properties
) {{
{I}{indent_but_first_line(body, I)}
}}"""
        ),
        None,
    )


@require(
    lambda cls: len(cls.concrete_descendants) > 0,
    "No dispatch possible without concrete descendants",
)
def _generate_dispatch_deserialize_implementation(
    cls: intermediate.ClassUnion,
) -> List[Stripped]:
    """Generate the impl. of the dispatching deserialization function for ``cls``."""
    targets: Iterable[intermediate.ConcreteClass]

    if isinstance(cls, intermediate.ConcreteClass):
        targets = itertools.chain([cls], cls.concrete_descendants)
    else:
        targets = cls.concrete_descendants

    assert targets is not None

    interface_name = cpp_naming.interface_name(cls.name)

    entries = []  # type: List[Stripped]
    for target_cls in targets:
        model_type = naming.json_model_type(target_cls.name)

        if len(target_cls.concrete_descendants) > 0:
            target_function = Stripped(
                cpp_naming.function_name(
                    Identifier(f"concretely_deserialize_{target_cls.name}")
                )
            )
        else:
            target_function = Stripped(
                cpp_naming.function_name(Identifier(f"deserialize_{target_cls.name}"))
            )

        target_function = Stripped(
            f"""\
{target_function}<
{I}types::{interface_name}
>"""
        )

        entries.append(
            Stripped(
                f"""\
{{
{I}{indent_but_first_line(cpp_common.string_literal(model_type), I)},
{I}{indent_but_first_line(target_function, I)}
}}"""
            )
        )

    entries_joined = ",\n".join(entries)

    dispatch_name = cpp_naming.constant_name(
        Identifier(f"deserialize_{cls.name}_by_model_type")
    )

    function_name = cpp_naming.function_name(Identifier(f"deserialize_{cls.name}"))

    return [
        Stripped(
            f"""\
std::map<
{I}std::string,
{I}std::function<
{II}std::pair<
{III}common::optional<std::shared_ptr<types::{interface_name}> >,
{III}common::optional<DeserializationError>
{II}>(const nlohmann::json&, bool)
{I}>
> {dispatch_name} = {{
{I}{indent_but_first_line(entries_joined, I)}
}};"""
        ),
        Stripped(
            f"""\
std::pair<
{I}common::optional<
{II}std::shared_ptr<types::{interface_name}>
{I}>,
{I}common::optional<DeserializationError>
> {function_name}(
{I}const nlohmann::json& json,
{I}bool additional_properties
) {{
{I}const std::string* model_type;
{I}common::optional<DeserializationError> error;

{I}std::tie(
{II}model_type,
{II}error
{I}) = GetModelTypeFrom(json);

{I}if (error.has_value()) {{
{II}return std::make_pair<
{III}common::optional<std::shared_ptr<types::{interface_name}> >,
{III}common::optional<DeserializationError>
{II}>(
{III}common::nullopt,
{III}std::move(error)
{II});
{I}}}

{I}const auto it = {dispatch_name}.find(*model_type);
{I}if (it == {dispatch_name}.end()) {{
{II}std::wstring message = common::Concat(
{III}L"The dispatch to the JSON de-serialization of "
{III}L"types::{interface_name} "
{III}L"is not defined for model type: ",
{III}common::Utf8ToWstring(*model_type)
{II});

{II}return std::make_pair<
{III}common::optional<std::shared_ptr<types::{interface_name}> >,
{III}common::optional<DeserializationError>
{II}>(
{III}common::nullopt,
{III}common::make_optional<DeserializationError>(
{IIII}message
{III})
{II});
{I}}}

{I}return (it->second)(json, additional_properties);
}}"""
        ),
    ]


def _generate_deserialization_implementation(cls: intermediate.ClassUnion) -> Stripped:
    """Generate the implementation of ``*From`` function."""
    deserialize_function = _determine_deserialize_function_to_call(cls=cls)

    interface_name = cpp_naming.interface_name(cls.name)

    deserialization_name = cpp_naming.function_name(Identifier(f"{cls.name}_from"))

    return Stripped(
        f"""\
common::expected<
{I}std::shared_ptr<types::{interface_name}>,
{I}DeserializationError
> {deserialization_name}(
{I}const nlohmann::json& json,
{I}bool additional_properties
) {{
{I}common::optional<
{II}std::shared_ptr<types::{interface_name}>
{I}> instance;

{I}common::optional<DeserializationError> error;

{I}std::tie(
{II}instance,
{II}error
{I}) = {indent_but_first_line(deserialize_function, I)}(
{II}json,
{II}additional_properties
{I});

{I}if (instance.has_value()) {{
{II}return std::move(*instance);
{I}}}

{I}if (!error.has_value()) {{
{II}throw std::logic_error(
{III}"Unexpected null error when null instance."
{II});
{I}}}
{I}return common::make_unexpected(
{II}std::move(*error)
{I});
}}"""
    )


def _generate_serialization_exception_implementation() -> List[Stripped]:
    """Generate the implementation of the ``SerializationException``."""
    return [
        Stripped("// region SerializationException"),
        Stripped(
            f"""\
std::string RenderSerializationErrorMessage(
{I}const std::wstring& cause,
{I}const iteration::Path& path
) {{
{I}return common::WstringToUtf8(
{II}common::Concat(
{III}L"Serialization failed at ",
{III}path.ToWstring(),
{III}L": ",
{III}cause
{II})
{I});
}}"""
        ),
        Stripped(
            f"""\
SerializationException::SerializationException(
{I}std::wstring cause,
{I}iteration::Path path
) :
{I}cause_(std::move(cause)),
{I}path_(std::move(path)),
{I}msg_(RenderSerializationErrorMessage(cause, path)) {{
{I}// Intentionally empty.
}}"""
        ),
        Stripped(
            f"""\
const char* SerializationException::what() const noexcept {{
{I}return msg_.c_str();
}}"""
        ),
        Stripped(
            f"""\
const std::wstring& SerializationException::cause() const noexcept {{
{I}return cause_;
}}"""
        ),
        Stripped(
            f"""\
const iteration::Path& SerializationException::path() const noexcept {{
{I}return path_;
}}"""
        ),
        Stripped("// endregion SerializationException"),
    ]


def _generate_serialize_int() -> Stripped:
    """Generate the function to serialize an integer to a JSON value."""
    return Stripped(
        f"""\
/**
 * \\brief Serialize the given number to a JSON value.
 *
 * We verify that the integer is within the range representable by 64-bit floats
 * for interoperability with other de-serializers.
 */
std::pair<
{I}common::optional<nlohmann::json>,
{I}common::optional<SerializationError>
> SerializeInt64(int64_t value) {{
{I}if (
{II}value < -9007199254740991L
{II}|| value > 9007199254740991L
{I}) {{
{II}const std::wstring message = common::Concat(
{III}L"The integer ",
{III}std::to_wstring(value),
{III}L" can not be serialized to JSON "
{III}L"as it is outside the range [-2^53 + 1, 2^53 - 1] and can not "
{III}L"be exactly represented as a 64-bit floating point number."
{II});

{II}return std::make_pair<
{III}common::optional<nlohmann::json>,
{III}common::optional<SerializationError>
{II}>(
{III}common::nullopt,
{III}common::make_optional<SerializationError>(
{IIII}message
{III})
{II});
{I}}}

{I}return std::make_pair<
{II}common::optional<nlohmann::json>,
{II}common::optional<SerializationError>
{I}>(
{II}common::make_optional<nlohmann::json>(value),
{II}common::nullopt
{I});
}}"""
    )


def _generate_serialize_str() -> Stripped:
    """Generate the function to serialize a wide string to a JSON value."""
    return Stripped(
        f"""\
/**
 * Serialize the given text to a JSON value.
 */
nlohmann::json SerializeWstring(
{I}const std::wstring& text
) {{
{I}return nlohmann::json(
{II}common::WstringToUtf8(text)
{I});
}}"""
    )


def _generate_serialize_bytearray() -> Stripped:
    """Generate the function to serialize a byte array to a JSON value."""
    return Stripped(
        f"""\
/**
 * Serialize the given bytes to a JSON value.
 */
nlohmann::json SerializeByteArray(
{I}const std::vector<std::uint8_t>& bytes
) {{
{I}return nlohmann::json(
{II}std::move(
{III}stringification::Base64Encode(bytes)
{II})
{I});
}}"""
    )


def _generate_serialize_iclass_definition() -> Stripped:
    """Generate the definition of the main dispatch for serializing ``IClass``."""
    return Stripped(
        f"""\
std::pair<
{I}common::optional<nlohmann::json>,
{I}common::optional<SerializationError>
> SerializeIClass(
{I}const types::IClass& that
);"""
    )


def _generate_serialize_primitive_property(
    getter_expr: str,
    primitive_type: intermediate.PrimitiveType,
    property_name: Identifier,
) -> Stripped:
    """
    Generate the snippet to serialize the given primitive property.

    The ``getter_expr`` refers to the C++ expression specifying the value
    to be serialized.

    The ``property_name`` refers to the intermediate property name.
    """
    json_prop_name_literal = cpp_common.string_literal(
        naming.json_property(property_name)
    )

    if primitive_type is intermediate.PrimitiveType.BOOL:
        return Stripped(f"result[{json_prop_name_literal}] = {getter_expr};")

    elif primitive_type is intermediate.PrimitiveType.INT:
        serialized_var = cpp_naming.variable_name(Identifier(f"json_{property_name}"))
        return Stripped(
            f"""\
common::optional<nlohmann::json> {serialized_var};
std::tie(
{I}{serialized_var},
{I}error
) = SerializeInt64(
{I}{indent_but_first_line(getter_expr, I)}
);
if (error.has_value()) {{
{I}error->path.segments.emplace_front(
{II}common::make_unique<iteration::PropertySegment>(
{III}iteration::Property::{cpp_naming.enum_literal_name(property_name)}
{II})
{I});

{I}return std::make_pair<
{II}common::optional<nlohmann::json>,
{II}common::optional<SerializationError>
{I}>(
{II}common::nullopt,
{II}std::move(error)
{I});
}}

result[{json_prop_name_literal}] = std::move(
{I}{serialized_var}
);"""
        )
    elif primitive_type is intermediate.PrimitiveType.FLOAT:
        return Stripped(
            f"""\
result[{json_prop_name_literal}] = {getter_expr};"""
        )

    elif primitive_type is intermediate.PrimitiveType.STR:
        serialized_var = cpp_naming.variable_name(Identifier(f"json_{property_name}"))
        return Stripped(
            f"""\
result[{json_prop_name_literal}] = SerializeWstring(
{I}{indent_but_first_line(getter_expr, II)}
);"""
        )

    elif primitive_type is intermediate.PrimitiveType.BYTEARRAY:
        return Stripped(
            f"""\
result[{json_prop_name_literal}] = stringification::Base64Encode(
{I}{getter_expr}
);"""
        )
    else:
        assert_never(primitive_type)


def _generate_serialize_property(prop: intermediate.Property) -> Stripped:
    """Generate the code snippet to serialize the property ``prop``."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    code = None  # type: Optional[Stripped]

    getter = cpp_naming.getter_name(prop.name)
    maybe_var = None
    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        maybe_var = cpp_naming.variable_name(Identifier(f"maybe_{prop.name}"))
        getter_expr = f"*{maybe_var}"
    else:
        getter_expr = f"that.{getter}()"

    json_prop_name = naming.json_property(prop.name)

    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
        code = _generate_serialize_primitive_property(
            getter_expr=getter_expr,
            primitive_type=type_anno.a_type,
            property_name=prop.name,
        )

    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        if isinstance(type_anno.our_type, intermediate.Enumeration):
            code = Stripped(
                f"""\
result[{cpp_common.string_literal(json_prop_name)}] = stringification::to_string(
{I}{indent_but_first_line(getter_expr, I)}
);"""
            )
        elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
            code = _generate_serialize_primitive_property(
                getter_expr=getter_expr,
                primitive_type=type_anno.our_type.constrainee,
                property_name=prop.name,
            )
        elif isinstance(
            type_anno.our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            serialized_var = cpp_naming.variable_name(Identifier(f"json_{prop.name}"))

            code = Stripped(
                f"""\
common::optional<nlohmann::json> {serialized_var};
std::tie(
{I}{serialized_var},
{I}error
) = SerializeIClass(
{I}*{indent_but_first_line(getter_expr, I)}
);
if (error.has_value()) {{
{I}error->path.segments.emplace_front(
{II}common::make_unique<iteration::PropertySegment>(
{III}iteration::Property::{cpp_naming.enum_literal_name(prop.name)}
{II})
{I});

{I}return std::make_pair<
{II}common::optional<nlohmann::json>,
{II}common::optional<SerializationError>
{I}>(
{II}common::nullopt,
{II}std::move(error)
{I});
}}

result[{cpp_common.string_literal(json_prop_name)}] = std::move(
{I}*{serialized_var}
);"""
            )
        else:
            assert_never(type_anno.our_type)
    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        assert isinstance(
            type_anno.items, intermediate.OurTypeAnnotation
        ) and isinstance(
            type_anno.items.our_type,
            (intermediate.AbstractClass, intermediate.ConcreteClass),
        ), (
            f"NOTE (mristin, 2023-11-21): We expect only lists of classes "
            f"at the moment, but you specified {type_anno}. "
            f"Please contact the developers if you need this feature."
        )

        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            assert maybe_var is not None
            size_expr = f"{maybe_var}->size()"
        else:
            size_expr = f"that.{getter}().size()"

        serialized_var = cpp_naming.variable_name(Identifier(f"json_{prop.name}"))

        const_ref_item_type = cpp_common.generate_type_with_const_ref_if_applicable(
            type_annotation=type_anno.items, types_namespace=cpp_common.TYPES_NAMESPACE
        )

        index_var = cpp_naming.variable_name(Identifier(f"index_{prop.name}"))
        code = Stripped(
            f"""\
nlohmann::json {serialized_var} = nlohmann::json::array();
{serialized_var}.get_ptr<nlohmann::json::array_t*>()->reserve(
{I}{indent_but_first_line(size_expr, I)}
);
size_t {index_var} = 0;
for (
{I}{const_ref_item_type} item
{I}: {getter_expr}
) {{
{I}common::optional<nlohmann::json> json_item;
{I}std::tie(
{II}json_item,
{II}error
{I}) = SerializeIClass(*item);

{I}if (error.has_value()) {{
{II}error->path.segments.emplace_front(
{III}common::make_unique<iteration::IndexSegment>(
{IIII}{index_var}
{III})
{II});

{II}error->path.segments.emplace_front(
{III}common::make_unique<iteration::PropertySegment>(
{IIII}iteration::Property::{cpp_naming.enum_literal_name(prop.name)}
{III})
{II});

{II}return std::make_pair<
{III}common::optional<nlohmann::json>,
{III}common::optional<SerializationError>
{II}>(
{III}common::nullopt,
{III}std::move(error)
{II});
{I}}}

{I}{serialized_var}.emplace_back(
{II}std::move(*json_item)
{I});

{I}++{index_var};
}}
result[{cpp_common.string_literal(json_prop_name)}] = std::move(
{I}{serialized_var}
);"""
        )
    else:
        assert_never(type_anno)

    assert code is not None

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        maybe_var_type = cpp_common.generate_type_with_const_ref_if_applicable(
            type_annotation=prop.type_annotation,
            types_namespace=cpp_common.TYPES_NAMESPACE,
        )

        assert maybe_var is not None

        code = Stripped(
            f"""\
{maybe_var_type} {maybe_var}(
{I}that.{getter}()
);
if ({maybe_var}.has_value()) {{
{I}{indent_but_first_line(code, I)}
}}"""
        )

    return code


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_serialize_cls(
    cls: intermediate.ConcreteClass,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the serialization function for the class ``cls``."""
    if cls.is_implementation_specific:
        implementation_key = specific_implementations.ImplementationKey(
            f"jsonization/serialize_{cls.name}.cpp"
        )

        code = spec_impls.get(implementation_key, None)
        if code is None:
            return None, Error(
                cls.parsed.node,
                f"The implementation is missing for the JSON serialization "
                f"of {cls.name!r}: {implementation_key}",
            )
        return code, None

    blocks = [
        Stripped(
            """\
nlohmann::json result = nlohmann::json::object();"""
        )
    ]  # type: List[Stripped]

    needs_error = False
    for prop in cls.properties:
        type_anno = intermediate.beneath_optional(prop.type_annotation)
        primitive_type = intermediate.try_primitive_type(type_anno)

        if primitive_type is not None and (
            primitive_type is intermediate.PrimitiveType.INT
        ):
            needs_error = True
            break

        if isinstance(type_anno, intermediate.OurTypeAnnotation) or (
            isinstance(type_anno, intermediate.ListTypeAnnotation)
            and isinstance(type_anno.items, intermediate.OurTypeAnnotation)
        ):
            needs_error = True
            break

    if needs_error:
        blocks.append(Stripped("common::optional<SerializationError> error;"))

    for prop in cls.properties:
        blocks.append(_generate_serialize_property(prop=prop))

    if cls.serialization.with_model_type:
        model_type_literal = cpp_common.string_literal(naming.json_model_type(cls.name))
        blocks.append(Stripped(f'result["modelType"] = {model_type_literal};'))

    blocks.append(
        Stripped(
            f"""\
return std::make_pair<
{I}common::optional<nlohmann::json>,
{I}common::optional<SerializationError>
>(
{I}common::make_optional<nlohmann::json>(std::move(result)),
{I}common::nullopt
);"""
        )
    )

    blocks_joined = "\n\n".join(blocks)

    serialize_name = cpp_naming.function_name(Identifier(f"serialize_{cls.name}"))

    interface_name = cpp_naming.interface_name(cls.name)

    return (
        Stripped(
            f"""\
std::pair<
{I}common::optional<nlohmann::json>,
{I}common::optional<SerializationError>
> {serialize_name}(
{I}const types::{interface_name}& that
) {{
{I}{indent_but_first_line(blocks_joined, I)}
}}"""
        ),
        None,
    )


def _generate_serialize_iclass_implementation(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the main dispatch function for serializing ``IClass``."""
    case_blocks = []  # type: List[Stripped]
    for cls in symbol_table.concrete_classes:
        serialize_name = cpp_naming.function_name(Identifier(f"serialize_{cls.name}"))

        model_type_literal = cpp_naming.enum_literal_name(cls.name)
        model_type_enum = cpp_naming.enum_name(Identifier("Model_type"))

        interface_name = cpp_naming.interface_name(cls.name)

        case_blocks.append(
            Stripped(
                f"""\
case types::{model_type_enum}::{model_type_literal}:
{I}return {serialize_name}(
{II}dynamic_cast<const types::{interface_name}&>(that)
{I});"""
            )
        )

    case_blocks.append(
        Stripped(
            f"""\
default: {{
{I}std::string message = common::Concat(
{II}"Unexpected model type: ",
{II}std::to_string(
{III}static_cast<std::uint32_t>(
{IIII}that.model_type()
{III})
{II})
{I});

{I}throw std::invalid_argument(message);
}}"""
        )
    )

    case_blocks_joined = "\n".join(case_blocks)

    return Stripped(
        f"""\
std::pair<
{I}common::optional<nlohmann::json>,
{I}common::optional<SerializationError>
> SerializeIClass(
{I}const types::IClass& that
) {{
{I}switch (that.model_type()) {{
{II}{indent_but_first_line(case_blocks_joined, II)}
{I}}};
}}"""
    )


def _generate_serialize_implementation() -> Stripped:
    """Generate the main serialization function."""
    return Stripped(
        f"""\
nlohmann::json Serialize(
{I}const types::IClass& that
) {{
{I}common::optional<nlohmann::json> result;
{I}common::optional<SerializationError> error;

{I}std::tie(
{II}result,
{II}error
{I}) = SerializeIClass(that);

{I}if (error.has_value()) {{
{II}throw SerializationException(
{III}std::move(error->cause),
{III}std::move(error->path)
{II});
{I}}}

{I}return std::move(*result);
}}"""
    )


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
    library_namespace: Stripped,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the C++ implementation of the de/serialization functions."""
    namespace = Stripped(f"{library_namespace}::{cpp_common.JSONIZATION_NAMESPACE}")

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "{include_prefix_path}/jsonization.hpp"
#include "{include_prefix_path}/stringification.hpp"
#include "{include_prefix_path}/wstringification.hpp"

#pragma warning(push, 0)
#include <functional>
#include <map>
#include <set>
#include <sstream>
#pragma warning(pop)"""
        ),
        cpp_common.generate_namespace_opening(namespace),
        *_generate_property_segment_implementation(),
        *_generate_index_segment_implementation(),
        *_generate_path_implementation(),
        Stripped("// region De-serialization"),
        *_generate_deserialization_error_implementation(),
        _generate_deserialize_bool(),
        _generate_deserialize_int(),
        _generate_deserialize_float(),
        _generate_deserialize_str(),
        _generate_deserialize_bytearray(),
        _generate_get_model_type(),
    ]

    for cls in symbol_table.classes:
        if isinstance(cls, intermediate.ConcreteClass):
            blocks.append(
                _generate_concretely_deserialize_definition(
                    cls=cls,
                )
            )

        if len(cls.concrete_descendants) > 0:
            blocks.append(_generate_dispatch_deserialize_definition(cls=cls))

    errors = []  # type: List[Error]

    for cls in symbol_table.classes:
        if isinstance(cls, intermediate.ConcreteClass):
            deserialize_block, error = _generate_concretely_deserialize_implementation(
                cls=cls,
                spec_impls=spec_impls,
            )
            if error is not None:
                errors.append(error)
                continue

            assert deserialize_block is not None
            blocks.append(deserialize_block)

        if len(cls.concrete_descendants) > 0:
            deserialize_dispatch_blocks = _generate_dispatch_deserialize_implementation(
                cls=cls
            )
            blocks.extend(deserialize_dispatch_blocks)

    for cls in symbol_table.classes:
        blocks.append(_generate_deserialization_implementation(cls=cls))

    blocks.extend(
        [
            Stripped("// endregion De-serialization"),
            Stripped("// region Serialization"),
            Stripped(
                f"""\
/**
 * \\brief Represent a serialization error.
 *
 * We use this error internally to avoid unnecessary stack unwinding,
 * but throw the \\ref SerializationException at the final site of
 * the serialization for the user.
 */
struct SerializationError {{
{I}/**
{I} * Human-readable description of the error
{I} */
{I}std::wstring cause;

{I}/**
{I} * Path to the value that caused the error
{I} */
{I}iteration::Path path;

{I}explicit SerializationError(
{II}std::wstring a_cause
{I}) : cause(std::move(a_cause)) {{
{II}// Intentionally empty.
{I}}}
}};  // struct SerializationError"""
            ),
            *_generate_serialization_exception_implementation(),
            _generate_serialize_int(),
            _generate_serialize_str(),
            _generate_serialize_bytearray(),
            _generate_serialize_iclass_definition(),
        ]
    )

    for cls in symbol_table.concrete_classes:
        serialize_block, error = _generate_serialize_cls(cls=cls, spec_impls=spec_impls)
        if error is not None:
            errors.append(error)
        else:
            assert serialize_block is not None
            blocks.append(serialize_block)

    blocks.append(_generate_serialize_iclass_implementation(symbol_table=symbol_table))

    blocks.append(_generate_serialize_implementation())

    blocks.extend(
        [
            Stripped("// endregion Serialization"),
            cpp_common.generate_namespace_closing(namespace),
            cpp_common.WARNING,
        ]
    )

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None



"""Generate the test code for enhancing the model instances."""

from typing import List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.csharp import common as csharp_common, naming as csharp_naming
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    namespace: csharp_common.NamespaceIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> str:
    """
    Generate the test code for enhancing the model instances.

    The ``namespace`` indicates the fully-qualified name of the base project.
    """
    blocks = [
        Stripped(
            f"""\
class Enhancement
{{
{I}public readonly long SomeCustomId;

{I}public Enhancement(long someCustomId)
{I}{{
{II}SomeCustomId = someCustomId;
{I}}}
}}"""
        ),
        Stripped(
            f"""\
private static AasEnhancing.Enhancer<Enhancement> CreateEnhancer()
{{
{I}long lastCustomId = 0;

{I}var enhancementFactory = new System.Func<IClass, Enhancement>(
{II}delegate
{II}{{
{III}lastCustomId++;
{III}return new Enhancement(lastCustomId);
{II}}}
{I});

{I}return new AasEnhancing.Enhancer<Enhancement>(enhancementFactory);
}}"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name = csharp_naming.class_name(concrete_cls.name)

        blocks.append(
            Stripped(
                f"""\
[Test]
public void Test_{cls_name}()
{{
{I}var instance = (
{II}Aas.Tests.CommonJsonization.LoadMaximal{cls_name}()
{I});

{I}var enhancer = CreateEnhancer();

{I}Assert.IsNull(enhancer.Unwrap(instance));

{I}var wrapped = enhancer.Wrap(instance);
{I}Assert.IsNotNull(wrapped);

{I}var idSet = new HashSet<long>();

{I}idSet.Add(enhancer.MustUnwrap(wrapped).SomeCustomId);
{I}idSet.UnionWith(
{II}wrapped
{III}.Descend()
{III}.Select(
{IIII}(descendant) =>
{IIIII}enhancer.MustUnwrap(descendant).SomeCustomId
{IIII})
{I});

{I}Assert.AreEqual(1, idSet.Min());
{I}Assert.AreEqual(idSet.Count, idSet.Max());
}}  // public void Test_{cls_name}"""
            )
        )

    blocks_joined = "\n\n".join(blocks)

    return f"""\
{csharp_common.WARNING}

using Aas = {namespace}; // renamed
using AasEnhancing = {namespace}.Enhancing; // renamed

using System.Collections.Generic; // can't alias
using System.Linq; // can't alias

using NUnit.Framework; // can't alias

namespace {namespace}.Tests
{{
{I}public class TestEnhancing
{I}{{
{II}{indent_but_first_line(blocks_joined, II)}
{I}}}  // class TestEnhancing
}}  // namespace {namespace}.Tests

{csharp_common.WARNING}
"""


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate C++ data structures to represent, iterate and transform an."""

from aas_core_codegen.cpp.structure import _generate

verify = _generate.verify
generate_header = _generate.generate_header
generate_implementation = _generate.generate_implementation



"""Generate the invariant verifiers from the intermediate representation."""

import io
import textwrap
from typing import (
    Tuple,
    Optional,
    List,
    Sequence,
    Mapping,
    Union,
)

from icontract import ensure, require

from aas_core_codegen import intermediate, specific_implementations
from aas_core_codegen.common import (
    Error,
    Stripped,
    assert_never,
    Identifier,
    indent_but_first_line,
    wrap_text_into_lines,
    assert_union_without_excluded,
)
from aas_core_codegen.intermediate import type_inference as intermediate_type_inference
from aas_core_codegen.parse import tree as parse_tree, retree as parse_retree
from aas_core_codegen.golang import (
    common as golang_common,
    naming as golang_naming,
    description as golang_description,
    pointering as golang_pointering,
    transpilation as golang_transpilation,
)
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


# region Verify


def verify(
    spec_impls: specific_implementations.SpecificImplementations,
    verification_functions: Sequence[intermediate.Verification],
) -> Optional[List[str]]:
    """Verify all the implementation snippets related to verification."""
    errors = []  # type: List[str]

    expected_keys = []  # type: List[specific_implementations.ImplementationKey]

    for func in verification_functions:
        if isinstance(func, intermediate.ImplementationSpecificVerification):
            expected_keys.append(
                specific_implementations.ImplementationKey(
                    f"Verification/{func.name}.go"
                ),
            )

    for key in expected_keys:
        if key not in spec_impls:
            errors.append(f"The implementation snippet is missing for: {key}")

    if len(errors) == 0:
        return None

    return errors


# endregion

# region Generate


class RegexRenderer(parse_retree.Renderer):
    """
    Render the regular expressions for Go.

    Notably, do not escape character points, but leave them as-are, since that is
    what Go regular expression engine expects.

    For example:

    .. code-block ::

        package main

        import (
            "fmt"
            "regexp"
        )

        func main() {
            re := regexp.MustCompile(
                "^[\x09\x0a\x0d\x20-\ud7ff\ue000-\ufffd\U00010000-\U0010ffff]*$",
            )
            text := "\U0001F600"
            fmt.Printf("%v", re.MatchString(text))
            // Prints "true"
        }

    """

    def char_to_str_and_escape_or_encode_if_necessary(
        self, node: parse_retree.Char, escaping: Mapping[str, str]
    ) -> List[Union[str, parse_tree.FormattedValue]]:
        if not node.explicitly_encoded:
            escaped = escaping.get(node.character, None)
            if escaped is not None:
                result: List[Union[str, parse_tree.FormattedValue]] = [escaped]
            else:
                result = [node.character]

            return result

        return [node.character]


_REGEX_RENDERER = RegexRenderer()


class _PatternVerificationTranspiler(
    parse_tree.RestrictedTransformer[Tuple[Optional[Stripped], Optional[Error]]]
):
    """Transpile a statement of a pattern verification into Golang."""

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def _transform_joined_str_values(
        self, values: Sequence[Union[str, parse_tree.FormattedValue]]
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        """Transform the values of a joined string to a Golang string literal."""
        # If we do not need interpolation, simply return the string literals
        # joined together.
        needs_interpolation = any(
            isinstance(value, parse_tree.FormattedValue) for value in values
        )
        if not needs_interpolation:
            return (
                Stripped(
                    golang_common.string_literal(
                        "".join(value for value in values)  # type: ignore
                    )
                ),
                None,
            )

        parts = []  # type: List[str]

        for value in values:
            if isinstance(value, str):
                parts.append(golang_common.string_literal(value))

            elif isinstance(value, parse_tree.FormattedValue):
                code, error = self.transform(value.value)
                if error is not None:
                    return None, error

                assert code is not None

                parts.append(code)
            else:
                assert_never(value)

        if len(parts) > 1:
            parts_joined = "\n".join(f"{part}," for part in parts)

            return (
                Stripped(
                    f"""\
aascommon.Concat(
{I}{indent_but_first_line(parts_joined, I)}
)"""
                ),
                None,
            )

        assert len(parts) == 1, "At least one part expected in the formatted string"
        return Stripped(parts[0]), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_constant(
        self, node: parse_tree.Constant
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if isinstance(node.value, str):
            # NOTE (mristin, 2023-04-12):
            # We assume that all the string constants are valid regular expressions.
            # At this point, we could not find any difference between Golang and
            # Python regex languages which are relevant to the features we currently
            # support.

            regex, parse_error = parse_retree.parse(values=[node.value])
            if parse_error is not None:
                regex_line, pointer_line = parse_retree.render_pointer(
                    parse_error.cursor
                )

                return (
                    None,
                    Error(
                        node.original_node,
                        f"The string constant could not be parsed "
                        f"as a regular expression: \n"
                        f"{parse_error.message}\n"
                        f"{regex_line}\n"
                        f"{pointer_line}",
                    ),
                )

            assert regex is not None

            # NOTE (mristin, 2022-11-04):
            # Strictly speaking, this is a joined string with a single value, a string
            # literal. Thus, do not be confused by the name of the function —
            # this function treats both joined formatted values *and* string literals.
            return self._transform_joined_str_values(
                values=parse_retree.render(regex=regex, renderer=_REGEX_RENDERER)
            )
        else:
            raise AssertionError(f"Unexpected {node=}")

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_name(
        self, node: parse_tree.Name
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        return Stripped(golang_naming.variable_name(node.identifier)), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_joined_str(
        self, node: parse_tree.JoinedStr
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        regex, parse_error = parse_retree.parse(values=node.values)
        if parse_error is not None:
            regex_line, pointer_line = parse_retree.render_pointer(parse_error.cursor)

            return (
                None,
                Error(
                    node.original_node,
                    f"The joined string could not be parsed "
                    f"as a regular expression: \n"
                    f"{parse_error.message}\n"
                    f"{regex_line}\n"
                    f"{pointer_line}",
                ),
            )

        assert regex is not None

        return self._transform_joined_str_values(
            values=parse_retree.render(regex=regex)
        )

    def transform_assignment(
        self, node: parse_tree.Assignment
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        assert isinstance(node.target, parse_tree.Name)
        variable = golang_naming.variable_name(node.target.identifier)
        code, error = self.transform(node.value)
        if error is not None:
            return None, error
        assert code is not None

        # NOTE (mristin, 2023-04-12):
        # We assume that the variables won't change in the patterns. If this assumption
        # is broken, fix the code here by first inspecting the scope and deciding
        # which variables need to be first defined and which have been already
        # defined.
        return Stripped(f"{variable} := {code}"), None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _transpile_pattern_verification(
    verification: intermediate.PatternVerification,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the verification function that checks the regular expressions."""
    # NOTE (mristin, 2022-11-12):
    # We assume that we performed all the checks at the intermediate stage.

    construct_name = golang_naming.private_function_name(
        Identifier(f"construct_{verification.name}")
    )

    blocks = []  # type: List[Stripped]

    # region Construct block

    writer = io.StringIO()
    writer.write(
        f"""\
func {construct_name}() *regexp.Regexp {{
"""
    )

    transpiler = _PatternVerificationTranspiler()

    for i, stmt in enumerate(verification.parsed.body):
        if i == len(verification.parsed.body) - 1:
            break

        code, error = transpiler.transform(stmt)
        if error is not None:
            return None, error
        assert code is not None

        writer.write(textwrap.indent(code, I))
        writer.write("\n")

    if len(verification.parsed.body) >= 2:
        writer.write("\n")

    pattern_expr, error = transpiler.transform(verification.pattern_expr)
    if error is not None:
        return None, error
    assert pattern_expr is not None

    writer.write(
        textwrap.indent(
            f"""\
return regexp.MustCompile(
{I}{pattern_expr},
)""",
            I,
        )
    )

    writer.write("\n}")

    blocks.append(Stripped(writer.getvalue()))

    # endregion

    # region Initialize the regex

    regex_name = golang_naming.private_constant_name(
        Identifier(f"{verification.name}_re")
    )

    blocks.append(Stripped(f"var {regex_name} = {construct_name}()"))

    # endregion

    # region Define the verification function

    assert len(verification.arguments) == 1
    assert isinstance(
        verification.arguments[0].type_annotation, intermediate.PrimitiveTypeAnnotation
    )
    # noinspection PyUnresolvedReferences
    assert (
        verification.arguments[0].type_annotation.a_type
        == intermediate.PrimitiveType.STR
    )

    arg_name = golang_naming.argument_name(verification.arguments[0].name)

    function_name = golang_naming.function_name(verification.name)

    writer = io.StringIO()

    if verification.description is not None:
        (comment, comment_errors,) = golang_description.generate_comment_for_signature(
            description=verification.description,
            context=golang_description.Context(
                package=golang_common.VERIFICATION_PACKAGE, cls_or_enum=None
            ),
        )
        if comment_errors is not None:
            return None, Error(
                verification.description.parsed.node,
                f"Failed to generate the documentation comment for {verification.name!r}",
                comment_errors,
            )

        assert comment is not None

        writer.write(comment)
        writer.write("\n")

    writer.write(
        f"""\
func {function_name}({arg_name} string) bool {{
{I}return {regex_name}.MatchString(
{II}{arg_name},
{I})
}}"""
    )

    blocks.append(Stripped(writer.getvalue()))

    # endregion

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    return Stripped(writer.getvalue()), None


class _TranspilableVerificationTranspiler(golang_transpilation.Transpiler):
    """Transpile the body of a :py:class:`TranspilableVerification`."""

    # fmt: off
    @require(
        lambda environment, verification:
        all(
            environment.find(arg.name) is not None
            for arg in verification.arguments
        ),
        "All arguments defined in the environment"
    )
    # fmt: on
    def __init__(
        self,
        type_map: Mapping[
            parse_tree.Node, intermediate_type_inference.TypeAnnotationUnion
        ],
        is_pointer_map: Mapping[parse_tree.Node, bool],
        environment: intermediate_type_inference.Environment,
        symbol_table: intermediate.SymbolTable,
        verification: intermediate.TranspilableVerification,
    ) -> None:
        """Initialize with the given values."""
        golang_transpilation.Transpiler.__init__(
            self,
            type_map=type_map,
            is_pointer_map=is_pointer_map,
            environment=environment,
        )

        self._symbol_table = symbol_table

        self._argument_name_set = frozenset(arg.name for arg in verification.arguments)

    def _transform_enumeration_literal(
        self, enumeration_name: Identifier, literal_name: Identifier
    ) -> Stripped:
        literal = golang_naming.enum_literal_name(
            enumeration_name=enumeration_name, literal_name=literal_name
        )
        return Stripped(f"aastypes.{literal}")

    def transform_name(
        self, node: parse_tree.Name
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if node.identifier in self._variable_name_set:
            return Stripped(golang_naming.variable_name(node.identifier)), None

        if node.identifier in self._argument_name_set:
            return Stripped(golang_naming.argument_name(node.identifier)), None

        if node.identifier in self._symbol_table.constants_by_name:
            constant_name = golang_naming.constant_name(node.identifier)
            return Stripped(f"aasconstants.{constant_name}"), None

        if node.identifier in self._symbol_table.verification_functions_by_name:
            return Stripped(golang_naming.function_name(node.identifier)), None

        our_type = self._symbol_table.find_our_type(name=node.identifier)
        if isinstance(our_type, intermediate.Enumeration):
            return (
                Stripped(f"aastypes.{golang_naming.enum_name(node.identifier)}"),
                None,
            )

        return None, Error(
            node.original_node,
            f"We can not determine how to transpile the name {node.identifier!r} "
            f"to Golang. We could not find it neither in the constants, nor in "
            f"verification functions, nor as an enumeration. "
            f"If you expect this name to be transpilable, please contact "
            f"the developers.",
        )


def _transpile_transpilable_verification(
    verification: intermediate.TranspilableVerification,
    symbol_table: intermediate.SymbolTable,
    environment: intermediate_type_inference.Environment,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Transpile a verification function."""
    # fmt: off
    type_inference, error = (
        intermediate_type_inference.infer_for_verification(
            verification=verification,
            base_environment=environment
        )
    )
    # fmt: on

    if error is not None:
        return None, error

    assert type_inference is not None

    pointer_inferrer = golang_pointering.Inferrer(
        environment=type_inference.environment_with_args,
        type_map=type_inference.type_map,
    )

    for node in verification.parsed.body:
        _ = pointer_inferrer.transform(node)

    if len(pointer_inferrer.errors) > 0:
        return None, Error(
            verification.parsed.node,
            f"Failed to infer whether a node is a Golang pointer "
            f"in the verification function {verification.name!r}",
            pointer_inferrer.errors,
        )

    transpiler = _TranspilableVerificationTranspiler(
        type_map=type_inference.type_map,
        is_pointer_map=pointer_inferrer.is_pointer_map,
        environment=type_inference.environment_with_args,
        symbol_table=symbol_table,
        verification=verification,
    )

    body = []  # type: List[Stripped]
    for node in verification.parsed.body:
        stmt, error = transpiler.transform(node)
        if error is not None:
            return None, Error(
                verification.parsed.node,
                f"Failed to transpile the verification function {verification.name!r}",
                [error],
            )

        assert stmt is not None
        body.append(stmt)

    writer = io.StringIO()

    if verification.description is not None:
        (comment, comment_errors,) = golang_description.generate_comment_for_signature(
            description=verification.description,
            context=golang_description.Context(
                package=golang_common.VERIFICATION_PACKAGE,
                cls_or_enum=None,
            ),
        )
        if comment_errors is not None:
            return None, Error(
                verification.description.parsed.node,
                f"Failed to generate the comment "
                f"for verification function {verification.name!r}",
                comment_errors,
            )

        assert comment is not None

        writer.write(comment)
        writer.write("\n")

    function_name = golang_naming.function_name(verification.name)

    if verification.returns is None:
        return_type_suffix = ""
    else:
        return_type = golang_common.generate_type(
            type_annotation=verification.returns, types_package=Identifier("aastypes")
        )
        return_type_suffix = f" {return_type}"

    arg_defs = []  # type: List[Stripped]
    for arg in verification.arguments:
        arg_type = golang_common.generate_type(
            arg.type_annotation, types_package=Identifier("aastypes")
        )
        arg_name = golang_naming.argument_name(arg.name)
        arg_defs.append(Stripped(f"{arg_name} {arg_type}"))

    if len(arg_defs) == 0:
        writer.write(
            f"""\
func {function_name}(){return_type_suffix} {{"""
        )
    else:
        arg_defs_joined = "\n".join(f"{arg_def}," for arg_def in arg_defs)
        writer.write(
            f"""\
func {function_name}(
{I}{indent_but_first_line(arg_defs_joined, I)}
){return_type_suffix} {{"""
        )

    if len(body) == 0:
        writer.write("\n")
        writer.write("// Intentionally empty.")
    else:
        for stmt in body:
            writer.write("\n")
            writer.write(textwrap.indent(stmt, I))

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


class _InvariantTranspiler(golang_transpilation.Transpiler):
    def __init__(
        self,
        type_map: Mapping[
            parse_tree.Node, intermediate_type_inference.TypeAnnotationUnion
        ],
        is_pointer_map: Mapping[parse_tree.Node, bool],
        environment: intermediate_type_inference.Environment,
        symbol_table: intermediate.SymbolTable,
    ) -> None:
        """Initialize with the given values."""
        golang_transpilation.Transpiler.__init__(
            self,
            type_map=type_map,
            is_pointer_map=is_pointer_map,
            environment=environment,
            types_package=Identifier("aastypes"),
        )

        self._symbol_table = symbol_table

    def _transform_enumeration_literal(
        self, enumeration_name: Identifier, literal_name: Identifier
    ) -> Stripped:
        literal = golang_naming.enum_literal_name(
            enumeration_name=enumeration_name, literal_name=literal_name
        )
        return Stripped(f"aastypes.{literal}")

    def transform_name(
        self, node: parse_tree.Name
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if node.identifier in self._variable_name_set:
            name = Stripped(golang_naming.variable_name(node.identifier))

        elif node.identifier == "self":
            # The ``that`` refers to the argument of the verification function.
            name = Stripped("that")

        elif node.identifier in self._symbol_table.constants_by_name:
            constant_name = golang_naming.constant_name(node.identifier)
            name = Stripped(f"aasconstants.{constant_name}")

        elif node.identifier in self._symbol_table.verification_functions_by_name:
            name = Stripped(golang_naming.function_name(node.identifier))

        elif (
            our_type := self._symbol_table.find_our_type(name=node.identifier),
            isinstance(our_type, intermediate.Enumeration),
        )[1]:
            name = Stripped(f"aastypes.{golang_naming.enum_name(node.identifier)}")
        else:
            return None, Error(
                node.original_node,
                f"We can not determine how to transpile the name {node.identifier!r} "
                f"to Golang. We could not find it neither in the local variables, "
                f"nor in the global constants, nor in verification functions, "
                f"nor as an enumeration. If you expect this name to be transpilable, "
                f"please contact the developers.",
            )

        assert name is not None
        return name, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _transpile_invariant(
    invariant: intermediate.Invariant,
    symbol_table: intermediate.SymbolTable,
    environment: intermediate_type_inference.Environment,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Translate the invariant from the meta-model into Golang code."""
    # fmt: off
    type_map, inference_error = (
        intermediate_type_inference.infer_for_invariant(
            invariant=invariant,
            environment=environment
        )
    )
    # fmt: on

    if inference_error is not None:
        return None, inference_error

    assert type_map is not None

    pointer_inferrer = golang_pointering.Inferrer(
        environment=environment, type_map=type_map
    )

    _ = pointer_inferrer.transform(invariant.body)

    if len(pointer_inferrer.errors) > 0:
        return None, Error(
            invariant.parsed.node,
            "Failed to infer whether a node is a Golang pointer in the invariant",
            pointer_inferrer.errors,
        )

    transpiler = _InvariantTranspiler(
        type_map=type_map,
        is_pointer_map=pointer_inferrer.is_pointer_map,
        environment=environment,
        symbol_table=symbol_table,
    )

    expr, error = transpiler.transform(invariant.parsed.body)
    if error is not None:
        return None, error

    assert expr is not None

    writer = io.StringIO()
    if len(expr) > 50 or "\n" in expr:
        writer.write(
            f"""\
if !(
{I}{indent_but_first_line(expr, I)}) {{
"""
        )
    else:
        no_parenthesis_type_in_this_context = (
            parse_tree.Index,
            parse_tree.Name,
            parse_tree.Member,
            parse_tree.MethodCall,
            parse_tree.FunctionCall,
        )

        if isinstance(invariant.parsed.body, no_parenthesis_type_in_this_context):
            not_expr = f"!{expr}"
        else:
            not_expr = f"!({expr})"

        writer.write(f"if {not_expr} {{\n")

    new_verification_error_writer = io.StringIO()

    new_verification_error_writer.write("newVerificationError(\n")

    # NOTE (mristin, 2023-04-12):
    # We need to wrap the description in multiple literals as a single long
    # string literal is often too much for the readability.
    invariant_description_lines = wrap_text_into_lines(invariant.description)

    if len(invariant_description_lines) == 1:
        line = invariant_description_lines[0]
        new_verification_error_writer.write(f"{I}{golang_common.string_literal(line)},")
        new_verification_error_writer.write(")")
    else:
        for i, line in enumerate(invariant_description_lines):
            if i == 0:
                new_verification_error_writer.write(
                    f"{I}{golang_common.string_literal(line)} +\n"
                )
            elif i < len(invariant_description_lines) - 1:
                new_verification_error_writer.write(
                    f"{I}{golang_common.string_literal(line)} +\n"
                )
            else:
                new_verification_error_writer.write(
                    f"{I}{golang_common.string_literal(line)},\n"
                )
                new_verification_error_writer.write(")")

    new_verification_error = Stripped(new_verification_error_writer.getvalue())

    writer.write(
        f"""\
{I}abort = onError(
{II}{indent_but_first_line(new_verification_error, II)},
{I})
{I}if abort {{
{II}return
{I}}}
}}"""
    )

    return Stripped(writer.getvalue()), None


OurTypeExceptEnumeration = Union[
    intermediate.ConstrainedPrimitive,
    intermediate.AbstractClass,
    intermediate.ConcreteClass,
]
assert_union_without_excluded(
    original_union=intermediate.OurType,
    subset_union=OurTypeExceptEnumeration,
    excluded=[intermediate.Enumeration],
)


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_verify_property_snippet(
    prop: intermediate.Property,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """
    Generate the snippet to verify a property.

    Return an empty string if there is nothing to be verified for the given property.
    """
    # NOTE (mristin, 2023-04-12):
    # Instead of writing here a complex but general solution with unrolling we choose
    # to provide a simple, but limited, solution. First, the meta-model is quite
    # limited itself at the moment, so the complexity of the general solution is not
    # warranted. Second, we hope that there will be fewer bugs in the simple solution
    # which is particularly important at this early adoption stage.
    #
    # We anticipate that in the future we will indeed need a general and complex
    # solution. Here are just some thoughts on how to approach it:
    # * Leave the pattern matching to produce more readable code for simple cases,
    # * Unroll only in case of composite types and optional composite types.

    type_anno = intermediate.beneath_optional(prop.type_annotation)

    prop_name = golang_naming.property_name(prop.name)
    prop_name_literal = golang_common.string_literal(prop_name)

    getter_name = golang_naming.getter_name(prop.name)

    optional = isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation)

    block = None  # type: Optional[Stripped]

    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
        if type_anno.a_type is intermediate.PrimitiveType.BYTEARRAY and not optional:
            block = Stripped(
                f"""\
if that.{getter_name}() == nil {{
{I}abort = onError(
{II}newVerificationError(
{III}"Required property not set: {prop_name}",
{II}),
{I})
{I}if abort {{
{II}return
{I}}}
}}"""
            )

    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        our_type = type_anno.our_type

        if isinstance(our_type, intermediate.Enumeration):
            enum_name = golang_naming.enum_name(our_type.name)

            # NOTE (mristin, 2023-04-12):
            # The case where we have no literals defined is an edge case where no
            # literal satisfies the condition.
            if len(our_type.literals) == 0:
                block = Stripped(
                    f"""\
err := newVerificationError(
{I}fmt.Sprintf(
{II}"The enumeration {enum_name} has no literals defined, " +
{II}"but you passed in: %v",
{II}that.{getter_name}()
{I})
)
err.Path.Prepend(
{I}&aasreporting.NameSegment{{
{II}Name: {prop_name_literal},
{I}}},
)
abort = onError(err)
if abort {{
{I}return
}}"""
                )
            else:
                first_literal = golang_naming.enum_literal_name(
                    enumeration_name=our_type.name,
                    literal_name=our_type.literals[0].name,
                )
                last_literal = golang_naming.enum_literal_name(
                    enumeration_name=our_type.name,
                    literal_name=our_type.literals[-1].name,
                )

                pointer_prefix = (
                    "*"
                    if golang_pointering.is_pointer_type(prop.type_annotation)
                    else ""
                )

                block = Stripped(
                    f"""\
if
{I}{pointer_prefix}that.{getter_name}() < aastypes.{first_literal} ||
{I}{pointer_prefix}that.{getter_name}() > aastypes.{last_literal} {{
{I}err := newVerificationError(
{II}fmt.Sprintf(
{III}"Invalid literal value for {enum_name}: %v",
{III}that.{getter_name}(),
{II}),
{I})
{I}err.Path.PrependName(
{II}&aasreporting.NameSegment{{
{III}Name: {prop_name_literal},
{II}}},
{I})
{I}abort = onError(err)
{I}if abort {{
{II}return
{I}}}
}}"""
                )

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            verify_function_name = golang_naming.function_name(
                Identifier(f"verify_{our_type.name}")
            )

            pointer_prefix = (
                "*" if golang_pointering.is_pointer_type(prop.type_annotation) else ""
            )

            block = Stripped(
                f"""\
abort = {verify_function_name}(
{I}{pointer_prefix}that.{getter_name}(),
{I}func(err *VerificationError) bool {{
{II}err.Path.PrependName(
{III}&aasreporting.NameSegment{{
{IIII}Name: {prop_name_literal},
{III}}},
{II})
{II}return onError(err)
{I}}},
)
if abort {{
{I}return
}}"""
            )

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            block = Stripped(
                f"""\
abort = Verify(
{I}that.{getter_name}(),
{I}func(err *VerificationError) bool {{
{II}err.Path.PrependName(
{III}&aasreporting.NameSegment{{
{IIII}Name: {prop_name_literal},
{III}}},
{II})
{II}return onError(err)
{I}}},
)
if abort {{
{I}return
}}"""
            )
        else:
            assert_never(our_type)

    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        assert isinstance(
            type_anno.items, intermediate.OurTypeAnnotation
        ) and isinstance(
            type_anno.items.our_type,
            (intermediate.AbstractClass, intermediate.ConcreteClass),
        ), (
            f"NOTE (mristin, 2023-04-12): We expect only lists of classes "
            f"at the moment, but you specified {type_anno}. "
            f"Please contact the developers if you need this feature."
        )

        block = Stripped(
            f"""\
for i, v := range that.{getter_name}() {{
{I}abort = Verify(
{II}v,
{II}func(err *VerificationError) bool {{
{III}err.Path.PrependIndex(
{IIII}&aasreporting.IndexSegment{{
{IIIII}Index: i,
{IIII}}},
{III})

{III}err.Path.PrependName(
{IIII}&aasreporting.NameSegment{{
{IIIII}Name: {prop_name_literal},
{IIII}}},
{III})

{III}return onError(err)
{II}}},
{I})
{I}if abort {{
{II}return
{I}}}
}}"""
        )

    else:
        assert_never(type_anno)

    primitive_type = intermediate.try_primitive_type(type_anno)

    is_reference = (
        optional
        or primitive_type is intermediate.PrimitiveType.BYTEARRAY
        or (
            isinstance(type_anno, intermediate.OurTypeAnnotation)
            and isinstance(
                type_anno.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            )
        )
        or isinstance(type_anno, intermediate.ListTypeAnnotation)
    )

    if not optional and is_reference and block is None:
        block = Stripped(
            f"""\
if that.{getter_name}() == nil {{
{I}abort = onError(
{II}newVerificationError(
{III}"Required property not set: {prop_name}",
{II}),
{I})
{I}if abort {{
{II}return
{I}}}
}}"""
        )
    elif not optional and is_reference and block is not None:
        block = Stripped(
            f"""\
if that.{getter_name}() == nil {{
{I}abort = onError(
{II}newVerificationError(
{III}"Required property not set: {prop_name}",
{II}),
{I})
{I}if abort {{
{II}return
{I}}}
}} else {{
{I}{indent_but_first_line(block, I)}
}}"""
        )
    elif optional and block is not None:
        block = Stripped(
            f"""\
if that.{getter_name}() != nil {{
{I}{indent_but_first_line(block, I)}
}}"""
        )
    elif block is None:
        return Stripped(""), None
    else:
        assert AssertionError(f"Unhandled case: {block=}, {optional=}, {is_reference=}")

    return block, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_verify_class(
    cls: intermediate.ConcreteClass,
    symbol_table: intermediate.SymbolTable,
    base_environment: intermediate_type_inference.Environment,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the verification function for the given concrete class."""
    errors = []  # type: List[Error]
    blocks = []  # type: List[Stripped]

    environment = intermediate_type_inference.MutableEnvironment(
        parent=base_environment
    )

    assert environment.find(Identifier("self")) is None
    environment.set(
        identifier=Identifier("self"),
        type_annotation=intermediate_type_inference.OurTypeAnnotation(our_type=cls),
    )

    # region Generate the non-recursive part verifying the invariants

    for invariant in cls.invariants:
        invariant_code, error = _transpile_invariant(
            invariant=invariant, symbol_table=symbol_table, environment=environment
        )
        if error is not None:
            errors.append(
                Error(
                    cls.parsed.node,
                    f"Failed to transpile the invariant of the class {cls.name!r}",
                    [error],
                )
            )
            continue

        assert invariant_code is not None

        blocks.append(invariant_code)

    # endregion

    # region Recurse into properties

    for prop in cls.properties:
        block, error = _generate_verify_property_snippet(prop=prop)
        if error is not None:
            errors.append(
                Error(
                    cls.parsed.node,
                    f"Failed to generate the verification of the property {prop.name!r} "
                    f"of the class {cls.name!r}",
                    [error],
                )
            )
            continue

        assert block is not None

        if block != "":
            blocks.append(block)

    # endregion

    if len(errors) > 0:
        return None, errors

    interface_name = golang_naming.interface_name(cls.name)

    if len(blocks) == 0:
        blocks.append(
            Stripped(
                f"""\
// No verification has been defined for {interface_name}."""
            )
        )

    function_name = golang_naming.function_name(Identifier(f"verify_{cls.name}"))

    body = "\n\n".join(blocks)

    return (
        Stripped(
            f"""\
// Verify `that` instance of [aastypes.{interface_name}].
//
// You have to supply the callback `onError` to iterate over the errors.
// If `onError` returns abort `true`, this function will abort
// further verification as well, and return abort `true`. Otherwise,
// abort `false` is returned.
func {function_name}(
{I}that aastypes.{interface_name},
{I}onError func(*VerificationError) bool,
) (abort bool) {{
{I}abort = false

{I}{indent_but_first_line(body, I)}

{I}return
}}"""
        ),
        None,
    )


def _generate_verify(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the main entry point for verification."""
    case_blocks = []  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        literal = golang_naming.enum_literal_name(
            enumeration_name=Identifier("Model_type"), literal_name=cls.name
        )

        verification_function = golang_naming.function_name(
            Identifier(f"verify_{cls.name}")
        )

        interface_name = golang_naming.interface_name(cls.name)

        case_blocks.append(
            Stripped(
                f"""\
case aastypes.{literal}:
{I}abort = {verification_function}(
{II}that.(aastypes.{interface_name}),
{II}onError,
{I})"""
            )
        )

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}abort = onError(
{II}newVerificationError(
{III}fmt.Sprintf(
{IIII}"Unexpected model type literal: %v",
{IIII}modelType,
{III}),
{II}),
{I})"""
        )
    )

    switch_body = Stripped("\n".join(case_blocks))
    switch_statement = Stripped(
        f"""\
switch modelType {{
{switch_body}
}}"""
    )

    model_type_getter = golang_naming.getter_name(Identifier("model_type"))

    return Stripped(
        f"""\
// Verify ``that`` instance.
//
// You have to supply the callback `onError` to iterate over the errors.
// If `onError` returns abort `true`, this function will abort
// further verification as well, and return abort `true`. Otherwise,
// abort `false` is returned.
func Verify(
{I}that aastypes.IClass,
{I}onError func(*VerificationError) bool,
) (abort bool) {{
{I}modelType := that.{model_type_getter}()
{I}{indent_but_first_line(switch_statement, I)}
{I}return
}}"""
    )


def _generate_verify_constrained_primitive(
    constrained_primitive: intermediate.ConstrainedPrimitive,
    symbol_table: intermediate.SymbolTable,
    base_environment: intermediate_type_inference.Environment,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the verify function for the constrained primitives."""
    errors = []  # type: List[Error]
    blocks = []  # type: List[Stripped]

    environment = intermediate_type_inference.MutableEnvironment(
        parent=base_environment
    )

    assert environment.find(Identifier("self")) is None
    environment.set(
        identifier=Identifier("self"),
        type_annotation=intermediate_type_inference.OurTypeAnnotation(
            our_type=constrained_primitive
        ),
    )

    for invariant in constrained_primitive.invariants:
        invariant_code, error = _transpile_invariant(
            invariant=invariant, symbol_table=symbol_table, environment=environment
        )
        if error is not None:
            errors.append(
                Error(
                    constrained_primitive.parsed.node,
                    f"Failed to transpile the invariant of "
                    f"the constrained primitive {constrained_primitive.name!r}",
                    [error],
                )
            )
            continue

        assert invariant_code is not None

        blocks.append(invariant_code)

    if len(errors) > 0:
        return None, errors

    if len(blocks) == 0:
        blocks.append(Stripped("// There is no verification specified."))

    body = "\n\n".join(blocks)

    function_name = golang_naming.function_name(
        Identifier(f"verify_{constrained_primitive.name}")
    )

    that_type = golang_common.PRIMITIVE_TYPE_MAP[constrained_primitive.constrainee]

    return (
        Stripped(
            f"""\
// Verify the constraints of `that` value.
//
// You have to supply the callback `onError` to iterate over the errors.
// If `onError` returns abort `true`, this function will abort
// further verification as well, and return abort `true`. Otherwise,
// abort `false` is returned.
func {function_name}(
{I}that {that_type},
{I}onError func(*VerificationError) bool,
) (abort bool) {{
{I}abort = false

{I}{indent_but_first_line(body, I)}

{I}return
}}"""
        ),
        None,
    )


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
    repo_url: Stripped,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the Golang code for verification based on the symbol table."""
    errors = []  # type: List[Error]

    aasconstants_url_literal = golang_common.string_literal(f"{repo_url}/constants")

    aascommon_url_literal = golang_common.string_literal(f"{repo_url}/common")

    aasreporting_url_literal = golang_common.string_literal(f"{repo_url}/reporting")

    aastypes_url_literal = golang_common.string_literal(f"{repo_url}/types")

    blocks = [
        Stripped(
            """\
// Package verification allows you to verify model instances.
//
// The main function is [Verify].
//
// Other verification functions (`Verify*`) are left for modularity, in case you want
// to be explicit about the typing in your code. However, in the large majority of
// the cases, you only want to call [Verify].
package verification"""
        ),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"math/big"
{I}"fmt"
{I}"regexp"
{I}"strconv"
{I}"strings"
{I}aascommon {aascommon_url_literal}
{I}aasconstants {aasconstants_url_literal}
{I}aasreporting {aasreporting_url_literal}
{I}aastypes {aastypes_url_literal}
)"""
        ),
        Stripped(
            f"""\
// Represent a verification violation.
//
// Implements `error`.
type VerificationError struct{{
{I}Path *aasreporting.Path
{I}Message string
}}"""
        ),
        Stripped(
            f"""\
func newVerificationError(message string) *VerificationError {{
{I}return &VerificationError{{
{II}Path: &aasreporting.Path{{}},
{II}Message: message,
{I}}}
}}"""
        ),
        Stripped(
            f"""\
func (ve *VerificationError) Error() string {{
{I}return fmt.Sprintf(
{II}"%s: %s",
{II}ve.PathString(),
{II}ve.Message,
{I})
}}"""
        ),
        Stripped(
            f"""\
// Render the path as a string.
func (ve *VerificationError) PathString() string {{
{I}return aasreporting.ToGolangPath(ve.Path)
}}"""
        ),
    ]  # type: List[Stripped]

    base_environment = intermediate_type_inference.populate_base_environment(
        symbol_table=symbol_table
    )

    for verification in symbol_table.verification_functions:
        if isinstance(verification, intermediate.ImplementationSpecificVerification):
            implementation_key = specific_implementations.ImplementationKey(
                f"Verification/{verification.name}.go"
            )

            implementation = spec_impls.get(implementation_key, None)
            if implementation is None:
                errors.append(
                    Error(
                        None,
                        f"The snippet for the verification function "
                        f"{verification.name!r} is missing: {implementation_key}",
                    )
                )
            else:
                blocks.append(implementation)

        elif isinstance(verification, intermediate.PatternVerification):
            implementation, error = _transpile_pattern_verification(
                verification=verification
            )

            if error is not None:
                errors.append(error)
            else:
                assert implementation is not None
                blocks.append(implementation)

        elif isinstance(verification, intermediate.TranspilableVerification):
            implementation, error = _transpile_transpilable_verification(
                verification=verification,
                symbol_table=symbol_table,
                environment=base_environment,
            )

            if error is not None:
                errors.append(error)
            else:
                assert implementation is not None
                blocks.append(implementation)

        else:
            assert_never(verification)

    for cls in symbol_table.concrete_classes:
        block, underlying_errors = _generate_verify_class(
            cls=cls, symbol_table=symbol_table, base_environment=base_environment
        )
        if underlying_errors is not None:
            errors.append(
                Error(
                    cls.parsed.node,
                    f"Failed to generate the verification for the class {cls.name!r}",
                    underlying_errors,
                )
            )
        else:
            assert block is not None
            blocks.append(block)

    for constrained_primitive in symbol_table.constrained_primitives:
        block, underlying_errors = _generate_verify_constrained_primitive(
            constrained_primitive=constrained_primitive,
            symbol_table=symbol_table,
            base_environment=base_environment,
        )
        if underlying_errors is not None:
            errors.append(
                Error(
                    constrained_primitive.parsed.node,
                    f"Failed to generate the verification for "
                    f"the constrained primitive {constrained_primitive.name!r}",
                    underlying_errors,
                )
            )
        else:
            assert block is not None
            blocks.append(block)

    blocks.append(_generate_verify(symbol_table=symbol_table))

    blocks.append(golang_common.WARNING)

    if len(errors) > 0:
        return None, errors

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None


# endregion
